from copy import deepcopy
from PyQt5 import QtWidgets
from PyQt5.QtWidgets import QCheckBox

try:
    from ..__db__ import SDB_DEFAULT
    from ..settings.base import VSpacer
except (ValueError, ImportError):
    from __db__ import SDB_DEFAULT
    from settings.base import VSpacer


class ResetDialog(QtWidgets.QDialog):
    def __init__(self, core):
        super().__init__()
        self.core = core
        self.resize(1, 1)
        self.setWindowTitle("Restore to defaults")
        buttonBox = QtWidgets.QDialogButtonBox()
        buttonBox.setStandardButtons(buttonBox.Cancel | buttonBox.Ok)
        buttonBox.setCenterButtons(True)
        buttonBox.accepted.connect(self.accept)
        buttonBox.rejected.connect(self.reject)

        self.checkboxes = {}
        layout = QtWidgets.QVBoxLayout(self)
        layout.addWidget(Settings(self))
        layout.addItem(VSpacer())
        layout.addWidget(buttonBox)

    def isChecked(self) -> bool:
        """Returns a boolean wether any checkbox is checked"""
        for key, item in self.checkboxes.items():
            if item.isChecked():
                return True
        return False

    def resetKey(self, key: str):
        """Restores default value of a top level database key"""
        self.core.sdb[key] = deepcopy(SDB_DEFAULT[key])

        if key == "general":  # Hook for menu icon color
            css = self.core.getNoteDecorationsCSS()
            self.core.setGlobalIconsColor("menu-icon", css["menu-icon"])
            self.core.setGlobalIconsColor("tray-icon", css["tray-icon"])

        elif key == "profile default":  # Hook for default name (general settings)
            default = SDB_DEFAULT["general"]["default name"]
            self.core.sdb["general"]["default name"] = default

        elif key == "mouse events":  # Hook for wheel treshold (general settings)
            default = SDB_DEFAULT["general"]["wheel threshold"]
            self.core.sdb["general"]["wheel threshold"] = default


class Settings(QtWidgets.QGroupBox):
    def __init__(self, parent):
        super().__init__("Please confirm which settings to reset", parent)
        parent.checkboxes["general"] = QCheckBox("General")
        parent.checkboxes["message bar"] = parent.checkboxes["general"]
        parent.checkboxes["launcher"] = parent.checkboxes["general"]
        parent.checkboxes["archives"] = QCheckBox("Logs and archives")
        parent.checkboxes["clean"] = parent.checkboxes["archives"]
        parent.checkboxes["terminal"] = QCheckBox("Advanced")
        parent.checkboxes["profile default"] = QCheckBox("Profile default")
        parent.checkboxes["key events"] = QCheckBox("Hotkeys")
        parent.checkboxes["mouse events"] = QCheckBox("Mouse events")
        parent.checkboxes["core menus"] = QCheckBox("Core menus")
        parent.checkboxes["context menus"] = QCheckBox("Note context menus")
        parent.checkboxes["toolbar menus"] = QCheckBox("Note toolbars menus")

        cbs = parent.checkboxes
        layout = QtWidgets.QGridLayout(self)
        layout.addWidget(self.AllCheckBox(cbs), 0, 0, 1, 1)
        layout.addWidget(cbs["core menus"], 0, 1, 1, 1)
        layout.addWidget(cbs["context menus"], 1, 1, 1, 1)
        layout.addWidget(cbs["general"], 1, 0, 1, 1)
        layout.addWidget(cbs["toolbar menus"], 2, 1, 1, 1)
        layout.addWidget(cbs["profile default"], 2, 0, 1, 1)
        layout.addWidget(cbs["archives"], 3, 1, 1, 1)
        layout.addWidget(cbs["key events"], 3, 0, 1, 1)
        layout.addWidget(cbs["terminal"], 4, 1, 1, 1)
        layout.addWidget(cbs["mouse events"], 4, 0, 1, 1)

    class AllCheckBox(QtWidgets.QCheckBox):
        def __init__(self, checkboxes: dict):
            super().__init__("All")
            self.checkboxes = checkboxes
            self.stateChanged.connect(self._update)

        def _update(self, state: int):
            """Propagates its own check state change to all checkboxes"""
            for key, item in self.checkboxes.items():
                item.setCheckState(state)



#!/usr/bin/python3
from typing import Callable


def callback(func: Callable) -> Callable:
    """Attaches a callback hook prior to function call"""

    def wrapper(self, *args, **kwargs) -> Callable:
        self.cb()
        return func(self, *args, **kwargs)

    return wrapper


class Observer(dict):
    """Dict object which emits a callback signal when its content, or its subitem content is modified"""

    def __init__(self, *args, callback: Callable):
        super().__init__(*args)
        self.cb = callback

    @callback
    def __delitem__(self, *args):
        super().__delitem__(*args)

    @callback
    def __setitem__(self, key: any, value: any):
        """Fires a callback function when a value is changed"""
        value = self._replace(value)
        super().__setitem__(key, value)
        if isinstance(value, dict):
            self._convert(value)

    def _convert(self, dct: dict):
        """Recursively converts nested dict to Observer subclasses"""
        for key, value in dct.items():
            dct[key] = self._replace(value)
            if isinstance(value, dict):
                self._convert(value)

    def _replace(self, obj: any) -> any:
        """Converts data structures to Observer subclasses"""
        obj = Observer(obj, callback=self.cb) if isinstance(obj, dict) else obj
        obj = ObserverSet(obj, callback=self.cb) if isinstance(obj, set) else obj
        obj = ObserverList(obj, callback=self.cb) if isinstance(obj, list) else obj
        return obj

    @callback
    def clear(self):
        super().clear()

    @callback
    def pop(self, *args) -> any:
        return super().pop(*args)

    @callback
    def popitem(self) -> any:
        return super().popitem()

    @callback
    def setdefault(self, *args) -> any:
        value = super().setdefault(*args)
        self._convert(self)
        return value

    @callback
    def update(self, *args):
        super().update(*args)
        self._convert(self)


class ObserverList(list):
    """List object which emits a callback signal when its content is modified"""

    def __init__(self, *args, callback: Callable):
        super().__init__(*args)
        self.cb = callback

    @callback
    def __delitem__(self, *args):
        super().__delitem__(*args)

    @callback
    def __iadd__(self, *args) -> list:
        return super().__iadd__(*args)

    @callback
    def __setitem__(self, *args):
        super().__setitem__(*args)

    @callback
    def append(self, *args):
        super().append(*args)

    @callback
    def clear(self):
        super().clear()

    @callback
    def extend(self, *args):
        super().extend(*args)

    @callback
    def insert(self, *args):
        super().insert(*args)

    @callback
    def pop(self, *args) -> any:
        return super().pop(*args)

    @callback
    def remove(self, *args):
        super().remove(*args)

    @callback
    def reverse(self):
        super().reverse()

    @callback
    def sort(self, *args, **kwargs):
        super().sort(*args, **kwargs)


class ObserverSet(set):
    """Set object which emits a callback signal when its content is modified"""

    def __init__(self, *args, callback: Callable):
        super().__init__(*args)
        self.cb = callback

    @callback
    def __iand__(self, *args) -> set:
        return super().__iand__(*args)

    @callback
    def __ior__(self, *args) -> set:
        return super().__ror__(*args)

    @callback
    def __isub__(self, *args) -> set:
        return super().__isub__(*args)

    @callback
    def __ixor__(self, *args) -> set:
        return super().__ixor__(*args)

    @callback
    def add(self, *args):
        super().add(*args)

    @callback
    def clear(self):
        super().clear()

    @callback
    def difference_update(self, *args):
        super().difference_update(*args)

    @callback
    def discard(self, *args):
        super().discard(*args)

    @callback
    def intersection_update(self, *args):
        super().intersection_update(*args)

    @callback
    def pop(self) -> any:
        return super().pop()

    @callback
    def remove(self, *args):
        super().remove(*args)

    @callback
    def symmetric_difference_update(self, *args):
        super().symmetric_difference_update(*args)

    @callback
    def update(self, *args):
        super().update(*args)



#!/usr/bin/python3
import sys
from pathlib import Path
from typing import Tuple
from PyQt5 import QtWidgets
from PyQt5.QtWidgets import QLabel, QSizePolicy

try:
    from ..backend import UserDirs, logger
    from ..backend.database import (
        AbstractDatabaseInterface,
        AutoCheckBox,
        AutoComboBoxGlob,
        AutoDoubleSpinBox,
        AutoLineEdit,
        AutoSpinBox,
    )
    from ..settings.base import HSpacer, VSpacer, Page, MODE_TO_KEY, KEY_TO_MODE
    from ..menus.core import ConfirmPrompt
except (ValueError, ImportError):
    from backend import UserDirs, logger
    from backend.database import (
        AbstractDatabaseInterface,
        AutoCheckBox,
        AutoComboBoxGlob,
        AutoDoubleSpinBox,
        AutoLineEdit,
        AutoSpinBox,
    )
    from settings.base import HSpacer, VSpacer, Page, MODE_TO_KEY, KEY_TO_MODE
    from menus.core import ConfirmPrompt

log = logger.new(__name__)


class PageProfile(Page):
    def __init__(self, core, db):
        super().__init__(core, db)
        self.core = core
        layout = QtWidgets.QVBoxLayout(self)
        layout.addWidget(Profile(self))
        layout.addItem(VSpacer())


class Profile(QtWidgets.QGroupBox):
    def __init__(self, parent):
        super().__init__("Default profile", parent)
        self.parent = parent
        self.core = parent.core
        styleBox = AutoComboBoxGlob(parent.db, ("profile default", "style"), "style")
        paletteBox = AutoComboBoxGlob(parent.db, ("profile default", "palette"), "palette")
        nameBox = AutoLineEdit(parent.db, ("general", "default name"))

        sizePolicy = QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(1)
        modeBox = ModeComboBox(parent.db, ("profile default", "mode"))
        opacityBox = AutoDoubleSpinBox(parent.db, ("profile default", "opacity"))
        wrapBox = AutoCheckBox("Word wrap", parent.db, ("profile default", "wrap"))
        pinBox = AutoCheckBox("Pinned", parent.db, ("profile default", "pin"))
        widthBox = AutoSpinBox(parent.db, ("profile default", "width"))
        widthBox.setSizePolicy(sizePolicy)
        widthBox.setSingleStep(10)
        heightBox = AutoSpinBox(parent.db, ("profile default", "height"))
        heightBox.setSizePolicy(sizePolicy)
        heightBox.setSingleStep(10)
        applyButton = QtWidgets.QPushButton("Apply to all")
        applyButton.clicked.connect(self._applyToAllNotes)

        self.setToolTip("Define the default options used for all\nnewly loaded notes")
        layout = QtWidgets.QGridLayout(self)
        layout.addWidget(QLabel("Size (w:h)"), 0, 0, 1, 1)
        layout.addWidget(widthBox, 0, 2, 1, 1)
        layout.addWidget(QLabel(":"), 0, 3, 1, 1)
        layout.addWidget(heightBox, 0, 4, 1, 1)
        layout.addItem(HSpacer(), 0, 5, 1, 1)
        layout.addWidget(QLabel("Opacity"), 1, 0, 1, 1)
        layout.addWidget(opacityBox, 1, 2, 1, 1)
        layout.addWidget(QLabel("Name"), 2, 0, 1, 1)
        layout.addWidget(nameBox, 2, 2, 1, 4)
        layout.addWidget(QLabel("Palette"), 3, 0, 1, 1)
        layout.addWidget(paletteBox, 3, 2, 1, 4)
        layout.addWidget(QLabel("Style"), 4, 0, 1, 1)
        layout.addWidget(styleBox, 4, 2, 1, 4)
        layout.addWidget(QLabel("Mode"), 5, 0, 1, 1)
        layout.addWidget(modeBox, 5, 2, 1, 4)
        layout.addWidget(wrapBox, 6, 0, 1, 3)
        layout.addWidget(pinBox, 7, 0, 1, 3)
        layout.addWidget(applyButton, 7, 3, 1, 3)

    def _applyToAllNotes(self):
        """Overwrites all notes profiles with the current default"""

        def _closeAllNotes():
            for path, note in dict(self.core.loaded).items():
                note.close()

        def _htmlToPlainText(path: Path):
            log.warning(f"Converting '{path}' from HTML to plain text ...")
            textEdit = QtWidgets.QTextEdit()
            with open(path, "r+", encoding="utf-8") as f:
                textEdit.setHtml(f.read())
                f.seek(0)
                f.write(textEdit.toPlainText())
                f.truncate()

        def _overwriteProfiles():
            default = self.parent.db["profile default"]
            log.debug(default)

            for key, profile in self.core.pdb.items():
                # Convert HTML to plain text
                path = UserDirs.NOTES / key
                if default["mode"] != "html" and path.is_html():
                    _htmlToPlainText(path)

                # Overwrite profile, preserve image mode
                mode = profile["mode"]
                profile.update(default)
                if mode == "image":
                    profile["mode"] = "image"

        message = "This will close all notes and\noverwrites their profiles. Continue ?"
        dialog = ConfirmPrompt("Confirmation", message)
        if dialog.exec_() == dialog.Yes:
            log.info("Overwriting all profiles with default settings ...")
            _closeAllNotes()
            _overwriteProfiles()


class ModeComboBox(AbstractDatabaseInterface, QtWidgets.QComboBox):
    def __init__(self, db: dict, slices: Tuple):
        super(QtWidgets.QComboBox, self).__init__()
        AbstractDatabaseInterface.__init__(self, db, slices)
        modes = ["Plain text", "Rich text"]
        if "QTermWidget" in sys.modules:
            modes += ["Console", "Vim"]

        self.addItems(sorted(modes))
        self.load()
        self.currentTextChanged.connect(self._save)

    def load(self):
        """Sets the default profile value as the current text"""
        mode = KEY_TO_MODE[self._load()]
        self.setCurrentText(mode)

    def value(self) -> str:
        """Translates the displayed mode label into its database key"""
        mode = MODE_TO_KEY[self.currentText()]
        return mode



#!/usr/bin/python3
from PyQt5 import QtWidgets
from PyQt5.QtWidgets import QLabel, QSizePolicy

try:
    from ..backend.database import AutoCheckBox, AutoSpinBox
    from ..settings.base import HSpacer, VSpacer, Page
except (ValueError, ImportError):
    from backend.database import AutoCheckBox, AutoSpinBox
    from settings.base import HSpacer, VSpacer, Page


class PageLogs(Page):
    def __init__(self, *args):
        super().__init__(*args)
        layout = QtWidgets.QVBoxLayout(self)
        layout.addWidget(General(self))
        layout.addWidget(Storage(self))
        layout.addItem(VSpacer())


class General(QtWidgets.QGroupBox):
    def __init__(self, parent):
        super().__init__("General", parent)
        sizePolicy = QSizePolicy(QSizePolicy.Maximum, QSizePolicy.Fixed)
        archiveBox = AutoSpinBox(parent.db, ("archives", "frequency"))
        archiveBox.setSizePolicy(sizePolicy)
        blankBox = AutoCheckBox("Remove blank notes on startup", parent.db, ("clean", "blanks"))
        archiveFormatText = AutoCheckBox("Text", parent.db, ("archives", "text"))
        archiveFormatImage = AutoCheckBox("Image", parent.db, ("archives", "image"))

        layout = QtWidgets.QGridLayout(self)
        layout.addWidget(blankBox, 0, 0, 1, 3)
        layout.addWidget(QLabel("Create an archive every"), 1, 0, 1, 1)
        layout.addWidget(archiveBox, 1, 1, 1, 1)
        layout.addWidget(QLabel("day(s)"), 1, 2, 1, 1)
        layout.addWidget(QLabel("Archive formats"), 2, 0, 1, 1)
        layout.addWidget(archiveFormatText, 2, 1, 1, 2)
        layout.addWidget(archiveFormatImage, 3, 1, 1, 2)


class Storage(QtWidgets.QGroupBox):
    def __init__(self, parent):
        super().__init__("Storage duration (days)", parent)
        trashBox = AutoSpinBox(parent.db, ("clean", "trash"))
        archiveBox = AutoSpinBox(parent.db, ("clean", "archives"))
        logBox = AutoSpinBox(parent.db, ("clean", "logs"))

        layout = QtWidgets.QGridLayout(self)
        layout.addItem(HSpacer(), 0, 2, 1, 1)
        layout.addWidget(QLabel("Archives"), 0, 0, 1, 1)
        layout.addWidget(archiveBox, 0, 1, 1, 1)
        layout.addWidget(QLabel("Trash"), 1, 0, 1, 1)
        layout.addWidget(trashBox, 1, 1, 1, 1)
        layout.addWidget(QLabel("Logs"), 2, 0, 1, 1)
        layout.addWidget(logBox, 2, 1, 1, 1)



#!/usr/bin/python3
import ctypes
import re
import sys
from cryptography.fernet import InvalidToken
from pathlib import Path
from typing import Iterable
from PyQt5 import QtGui, QtCore, QtWidgets
from PyQt5.QtWidgets import QLabel, QSizePolicy, QSpacerItem

try:
    from ..__id__ import APP_NAME, ID
    from ..backend import UserFiles, UserDirs, RootDirs, logger, sanitizeFileName
    from ..backend.database import AutoCheckBox, AutoComboBoxGlob, AutoFontComboBox, AutoSpinBox
    from ..backend.cryptography import Cipher, MasterKeyPrompt, PasswordInputDialog
    from ..settings.base import HSpacer, VSpacer, Page
except (ValueError, ImportError):
    from __id__ import APP_NAME, ID
    from backend import UserFiles, UserDirs, RootDirs, logger, sanitizeFileName
    from backend.database import AutoCheckBox, AutoComboBoxGlob, AutoFontComboBox, AutoSpinBox
    from backend.cryptography import Cipher, MasterKeyPrompt, PasswordInputDialog
    from settings.base import HSpacer, VSpacer, Page

log = logger.new(__name__)


class PageAdvanced(Page):
    def __init__(self, *args):
        super().__init__(*args)
        layout = QtWidgets.QVBoxLayout(self)
        layout.addWidget(Terminal(self))
        layout.addWidget(Encryption(self))
        layout.addWidget(Symlink(self))
        layout.addWidget(Folder(self))
        layout.addItem(VSpacer())


class Terminal(QtWidgets.QGroupBox):
    def __init__(self, parent):
        super().__init__("Terminal", parent)
        colorSchemeBox = AutoComboBoxGlob(parent.db, ("terminal", "color scheme"), "terminal")
        fontSizeBox = AutoSpinBox(parent.db, ("terminal", "font size"))
        fontFamilyBox = AutoFontComboBox(parent.db, ("terminal", "font family"))

        layout = QtWidgets.QGridLayout(self)
        layout.addWidget(QLabel("Colorscheme"), 0, 0, 1, 1)
        layout.addWidget(colorSchemeBox, 0, 1, 1, 2)
        layout.addWidget(QLabel("Font family"), 1, 0, 1, 1)
        layout.addWidget(fontFamilyBox, 1, 1, 1, 2)
        layout.addWidget(QLabel("Font size"), 2, 0, 1, 1)
        layout.addWidget(fontSizeBox, 2, 1, 1, 1)
        layout.addItem(HSpacer(), 2, 2, 1, 1)


class Encryption(QtWidgets.QGroupBox):
    def __init__(self, parent):
        super().__init__("Encryption", parent)
        self.core = parent.core
        setButton = QtWidgets.QPushButton("Set/Replace Master Key")
        setButton.clicked.connect(self._setPassword)
        encryptAllBox = AutoCheckBox("Encrypt all notes by default", parent.db, ("general", "encrypt all"))
        self.removeButton = QtWidgets.QPushButton("Remove Master Key")
        self.removeButton.clicked.connect(self._removePassword)
        self.removeButton.setEnabled(UserFiles.KEYRING.is_file())

        layout = QtWidgets.QGridLayout(self)
        layout.addWidget(setButton, 1, 0, 1, 1)
        layout.addWidget(self.removeButton, 1, 1, 1, 1)
        layout.addWidget(encryptAllBox, 0, 0, 1, 2)

    def _findEncrypted(self) -> Iterable:
        """Iterates over text files encrypted with the master key"""
        files = [f for f in UserDirs.NOTES.rglob("*.txt") if f.is_file()]
        for path in files:
            with open(path, encoding="utf-8") as f:
                content = f.read()
            if not content and path in self.core.loaded:
                if self.core.loaded[path].key == self.core.masterKey:
                    yield path, ""
            elif content.startswith(Cipher.MAGIC):
                length = len(Cipher.MAGIC)
                token = content[length:].encode()
                try:
                    yield path, Cipher.decrypt(self.core.masterKey, token)
                except InvalidToken:
                    pass

    def _removePassword(self):
        """Removes encryption of all master key encrypted files"""
        if not self.core.masterKey:
            if MasterKeyPrompt.login(self.core):
                self._removePassword()
            return

        for path, text in self._findEncrypted():
            if path in self.core.loaded:
                self.core.loaded[path].content = text
                self.core.loaded[path].keyring.setKey(None)
            else:
                with open(path, "w", encoding="utf-8") as f:
                    f.write(text)
            log.info(f"{path.name} : Removed encryption")
        UserFiles.KEYRING.unlink()
        self.core.masterKey = None
        self.removeButton.setEnabled(False)
        log.info(f"Removed master token '{UserFiles.KEYRING}'")

    def _replacePassword(self):
        """Replaces the password of all master key encrypted files"""
        dialog = PasswordInputDialog(
            self.core, "Master key replacement", "Please provide a new master key."
        )
        if dialog.exec_() == dialog.Accepted:
            for path, text in self._findEncrypted():
                if path in self.core.loaded:
                    self.core.loaded[path].keyring.setKey(dialog.key)
                else:
                    with open(path, "w", encoding="utf-8") as f:
                        token = Cipher.encrypt(dialog.key, text)
                        token = Cipher.MAGIC + token.decode("utf-8")
                        f.write(token)
                log.info(f"{path.name} : Replaced key")
            MasterKeyPrompt.save(self.core, dialog.key)

    def _setPassword(self):
        """Sets or replaces the master key"""
        if not UserFiles.KEYRING.is_file():
            if MasterKeyPrompt.set(self.core):
                self.removeButton.setEnabled(True)

        elif not self.core.masterKey:
            if MasterKeyPrompt.login(self.core):
                self._setPassword()

        else:
            self._replacePassword()


class Folder(QtWidgets.QGroupBox):
    def __init__(self, parent):
        super().__init__("Configuration files", parent)
        self.core = parent.core

        sizePolicy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Maximum)
        button = QtWidgets.QPushButton(f"Edit {APP_NAME}\nconfiguration")
        button.setSizePolicy(sizePolicy)
        button.setFocusPolicy(QtCore.Qt.NoFocus)
        button.clicked.connect(self._clicked)

        sizePolicy = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Maximum)
        label = QLabel("Manually edit CSS, Vim and\nQTermWidget configuration files")
        label.setSizePolicy(sizePolicy)
        label.setWordWrap(True)

        spacer = QSpacerItem(5, 0, QSizePolicy.Fixed, QSizePolicy.Maximum)
        layout = QtWidgets.QHBoxLayout(self)
        layout.addWidget(label)
        layout.addItem(spacer)
        layout.addWidget(button)

    def _clicked(self):
        self.core.fileManager(UserDirs.CONFIG / "ui")


class Symlink(QtWidgets.QGroupBox):
    def __init__(self, parent):
        super().__init__("Symbolic links", parent)
        self.core = parent.core

        sizePolicy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Maximum)
        button = QtWidgets.QPushButton("Create symbolic\nlinks (symlinks)")
        button.setSizePolicy(sizePolicy)
        button.setFocusPolicy(QtCore.Qt.NoFocus)
        button.clicked.connect(self._clicked)
        button.setToolTip("Add new symbolic links to the notes repository")

        sizePolicy = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Maximum)
        label = QLabel(
            "This utility helps to create symbolic links and add them into the note repository. "
            "Use this tool to link a directory into the repository (ie. cloud storage drive)."
        )
        label.setSizePolicy(sizePolicy)
        label.setWordWrap(True)

        spacer = QSpacerItem(5, 0, QSizePolicy.Fixed, QSizePolicy.Maximum)
        layout = QtWidgets.QHBoxLayout(self)
        layout.addWidget(label)
        layout.addItem(spacer)
        layout.addWidget(button)

    def _clicked(self):
        SymlinkDialog().exec()


class SymlinkDialog(QtWidgets.QDialog):
    def __init__(self):
        super().__init__()
        icon = QtGui.QIcon(str(RootDirs.ICONS / f"{ID}_settings.svg"))
        self.resize(420, 1)
        self.setWindowTitle("Symlinks Creation Utility")
        self.setWindowIcon(icon)

        self.browseButton = QtWidgets.QPushButton("Browse")
        self.browseButton.clicked.connect(self._browse)
        self.createButton = QtWidgets.QPushButton("Create symlink")
        self.createButton.clicked.connect(self._createSymlink)
        self.createButton.setEnabled(False)

        self.msgLabel = QLabel()
        self.targetLine = QtWidgets.QLineEdit()
        self.destLine = QtWidgets.QLineEdit()
        self.targetLine.textChanged.connect(self._nameChanged)
        self.destLine.textChanged.connect(self._nameChanged)

        buttonBox = QtWidgets.QDialogButtonBox()
        buttonBox.setStandardButtons(buttonBox.Close)
        buttonBox.rejected.connect(self.close)

        spacer = QSpacerItem(0, 10, QSizePolicy.Maximum, QSizePolicy.Fixed)
        gridLayout = QtWidgets.QGridLayout(self)
        gridLayout.addWidget(QLabel("Target"), 0, 0, 1, 1)
        gridLayout.addWidget(self.targetLine, 0, 1, 1, 1)
        gridLayout.addWidget(self.browseButton, 0, 2, 1, 1)
        gridLayout.addWidget(QLabel("Name"), 1, 0, 1, 1)
        gridLayout.addWidget(self.destLine, 1, 1, 1, 1)
        gridLayout.addWidget(self.createButton, 1, 2, 1, 1)
        gridLayout.addItem(spacer, 2, 0, 1, 1)
        gridLayout.addWidget(self.msgLabel, 2, 1, 1, 1)
        gridLayout.addWidget(buttonBox, 3, 0, 1, 3)
        self.browseButton.setFocus()

    @property
    def destination(self) -> Path:
        sep = "\\" if sys.platform.startswith("win") else "/"
        name = self.destLine.text().split(sep)
        return Path(UserDirs.NOTES, *name)

    @property
    def target(self) -> Path:
        return Path(self.targetLine.text())

    def _assertSymlink(self):
        """Shows a sucess/failure message following symlink creation"""
        self._block(False)
        if self.destination.is_symlink():
            self.targetLine.clear()
            self.destLine.clear()
            self._setMessage(f"Symlink successfully created:\n{self.destination}", "black")
        else:
            self._setMessage(f"Symlink creation failed:\n{self.destination}", "red")

    def _block(self, enabled: bool):
        """Toggles interactive widgets"""
        self.targetLine.setEnabled(not enabled)
        self.destLine.setEnabled(not enabled)
        self.browseButton.setEnabled(not enabled)
        self.createButton.setEnabled(not enabled)

    def _browse(self):
        """Opens a folder selection dialog"""
        dialog = self.BrowseDialog()
        if dialog.exec_() == dialog.Accepted:
            dest = dialog.selectedFiles()[0]
            self.targetLine.setText(dest)

    def _createSymlink(self):
        """Creates a symbolic link, ask for user account control (UAC) elevation in Windows"""
        self._block(True)
        try:
            args = f'/c mklink /D "{self.destination}" "{self.target}"'
            ctypes.windll.shell32.ShellExecuteW(None, "runas", "cmd", args, None, 1)
        except AttributeError:
            self.destination.symlink_to(self.target)
        QtCore.QTimer.singleShot(1000, self._assertSymlink)

    def _isValidName(self) -> bool:
        """Compares link's name with the valid name filter used in move/rename"""
        name = re.sub(r"[\:\/\\]", "", self.destLine.text())
        return name == sanitizeFileName(self.destLine.text())

    def _nameChanged(self, text: str):
        """Updates error message on change"""
        self.createButton.setEnabled(False)

        if not self.destLine.text() or not self.targetLine.text():
            self.msgLabel.clear()

        elif not self._isValidName():
            self._setMessage("Illegal name or characters", "red")

        elif not self.target.exists():
            self._setMessage("Target directory does not exist", "red")

        elif self.destination.exists():
            self._setMessage("Destination folder already exist", "red")

        else:
            self.msgLabel.clear()
            self.createButton.setEnabled(True)

    def _setMessage(self, text: str, color: str):
        self.msgLabel.setStyleSheet(f"color: {color};")
        self.msgLabel.setText(text)

    class BrowseDialog(QtWidgets.QFileDialog):
        def __init__(self):
            super().__init__()
            home = str(Path.home())
            self.setDirectory(home)
            self.setWindowTitle("Select a target directory")
            self.setFileMode(self.Directory)
            self.setOptions(self.ShowDirsOnly)
            self.setAcceptMode(self.AcceptOpen)
            self.setModal(True)



#!/usr/bin/python3
import re
from collections import namedtuple
from typing import Iterator
from PyQt5 import QtWidgets
from PyQt5.QtWidgets import QSizePolicy, QSpacerItem

try:
    from ..notes import NoteSetup
except (ValueError, ImportError):
    from notes import NoteSetup

ActionID = namedtuple("ActionID", ("key", "type", "icon"))
KEY_TO_MODE = {
    "console": "Console",
    "image": "Image",
    "html": "Rich text",
    "plain": "Plain text",
    "vim": "Vim",
}
MODE_TO_KEY = {v: k for k, v in KEY_TO_MODE.items()}


class VSpacer(QSpacerItem):
    def __init__(self, height: int = 40):
        policy = QSizePolicy.Minimum, QSizePolicy.Expanding
        super().__init__(0, height, *policy)


class HSpacer(QSpacerItem):
    def __init__(self, width: int = 40):
        policy = QSizePolicy.Expanding, QSizePolicy.Minimum
        super().__init__(width, 0, *policy)


class Page(QtWidgets.QWidget):
    def __init__(self, core, db: dict):
        super().__init__()
        self.core = core
        self.db = db


class ActionBox(QtWidgets.QComboBox):
    def __init__(self, core):
        super().__init__()
        self.core = core
        self.setEditable(True)
        self.setInsertPolicy(QtWidgets.QComboBox.NoInsert)
        self.completer().setCompletionMode(QtWidgets.QCompleter.PopupCompletion)

    def noteActionsSetEnabled(self, enabled: bool):
        """Toggles enabled status of notes-only actions"""
        for i in range(self.count()):
            item = self.model().item(i)
            if enabled:
                item.setEnabled(True)
            elif item.text().endswith(" (note)"):
                item.setEnabled(False)

    def populate(self):
        """Fetch, sort and fill the combo box with all actions"""
        exclusions = ("folders list", "notes list", "separator")
        sets = {"note": NoteSetup.actionsFromMode(self.core), "core": self.core.actions.tray}
        for type_, actions in sets.items():
            actions = [f"{actions[a].label} ({type_})" for a in actions if a not in exclusions]
            self.addItems(sorted(actions))


class ActionTree(QtWidgets.QTreeWidget):
    def __init__(self, parent):
        super().__init__()
        self.core = parent.core
        self.db = parent.db
        self.setIndentation(20)
        self.setAlternatingRowColors(True)

    def iterate(self) -> Iterator[QtWidgets.QTreeWidgetItem]:
        """Iterator function for the top level items"""
        for i in range(self.topLevelItemCount()):
            yield self.topLevelItem(i)

    def labelToAction(self, text: str) -> ActionID:
        """Converts an action label into a tuple of action identifying data"""
        label, type_ = re.findall(r"(.+)\s+\((\w+)\)$", text)[0]
        for actions in (NoteSetup.actionsFromMode(self.core), self.core.actions.tray):
            for key, action in actions.items():
                if label == action.label:
                    return ActionID(key, type_, action.icon)



VERSION = 1.5
DATE = "2024/03/22"

APP_NAME = "PrimeNote"
ID = APP_NAME.lower()
DESC_AUTHOR = "© William Belanger, licensed under GPLv3"
DESC_SHORT = f"{APP_NAME} is a free and open source note-taking tool."
DESC_LONG = f"{APP_NAME} is a clever and highly adaptable note-editing tool designed for both Linux and Windows. Its functionality allows for swift editing, storage, organization, and backup of an unlimited number of notes. With {APP_NAME}, users can take advantage of advanced text operations and an extensive range of modes encompassing text and image processing, as well as a built-in terminal emulator and Vim integration for advanced users."

LINK_AUR = "https://aur.archlinux.org/packages/primenote-git"
LINK_DONATION = "https://www.paypal.com/donate?hosted_button_id=7UTK3HPH6Q5DG"
LINK_HOMEPAGE = "https://www.primenote.cc"
LINK_SOURCE = "https://www.gitlab.com/william.belanger/primenote"

HELP = [
    "\033[1m--help, -h:\033[0m Show this help message",
    "\033[1m--reset, -r <option>:\033[0m Reset configuration files to default",
    "\tOptions: {all,console,css,profiles,settings,vim}",
    "\tDefault: {all}\n",
    "\033[1m--search, -s <option>:\033[0m Toggle, show or hide the note search utility",
    "\tOptions: {toggle,show,hide}",
    "\tDefault: {toggle}\n",
    "\033[1m--verbose, -v <option>:\033[0m Set console logging level",
    "\tOptions: {debug,info,warning,error,critical,quiet}",
    "",
    "\033[1mCORE ACTIONS\033[0m",
    "\033[1m--core, -c <action>\033[0m",
    "  hide\tHide all notes",
    "  load\tLoad all notes in nested folders",
    "  new\tCreate a new note",
    "  menu\tPopup main menu at mouse position",
    "  open\tOpen notes and settings folder in file manager\t",
    "  quit\tClose the application",
    "  reset\tShow all and reset profiles",
    "  reverse\tToggle visibility of loaded notes",
    "  search\tToggle the note search utility",
    "  settings\tOpen application settings",
    "  show\tShow all notes",
    "  toggle\tToggle visibility of favorites notes",
    "  unload\tUnload all notes in nested folders",
    "  wizard\tOpen the setup wizard",
    "",
    "\033[1mNOTE ACTIONS\033[0m",
    "\033[1m--note, -n <action> --path, -p <path>\033[0m",
    "  activate\tRaises and focus the window (void ::activateWindow)",
    "  delete\tDelete note",
    "  duplicate\tDuplicate note",
    "  export\tOpen an export browser dialog",
    "  hide\tHide note (void ::hide)",
    "  move\tMove note in a new folder",
    "  opacity+\tIncrease window opacity",
    "  opacity-\tDecrease window opacity",
    "  open\tOpen note directory in file manager",
    "  pin\tToggle pin status",
    "  raise\tRaises the window (void ::raise)",
    "  refresh\tUpdate the window icons and stylesheets",
    "  rename\tOpen rename dialog",
    "  roll\tMinimize screen space used, leaving only the titlebar",
    "  show\tShow note",
    "  unload\tUnload a note if nested, hide if at root",
    "  unload\tUnload nested note or hide root note",
    "  unroll\tRestore a note to its full size",
    "",
    "\033[1mNOTE ACTIONS – TEXT MODE\033[0m",
    "  copy\tCopy plain text to clipboard",
    "  cut\tDelete selection and copy plain text to clipboard",
    "  line delete\tDelete selection with its linebreak",
    "  line down\tShift line position down",
    "  line duplicate\tDuplicate current line",
    "  line end\tMove cursor position at end of the line",
    "  line start\tMove cursor position at beginning of the line",
    "  line up\tShift line position up",
    "  lock\tSet file permissions to read-only",
    "  lock|unlock\tToggle read-only mode",
    "  lowercase\tSet selection to lowercase",
    "  paste\tInsert clipboard content",
    "  redo\tRedo the last operation (void ::redo)",
    "  search\tOpen the find and replace utility",
    "  select all\tSelect all text (void ::selectAll)",
    "  shuffle\tShuffle selection",
    "  save\tSave changes to file",
    "  sort\tSort selection",
    "  special paste\tPaste text while replacing linebreaks",
    "  unlock\tSet file permissions to read-write",
    "  undo\tUndo the last operation (void ::undo)",
    "  uppercase\tSet selection to uppercase",
    "  wrap\tToggle word-wrap",
    "",
    "\033[1mNOTE ACTIONS – RICH TEXT MODE\033[0m",
    "  bold\tToggle bold decoration on the anchor or the selected text",
    "  clear format\tClear all text formatting",
    "  copy rich\tCopy formatted text to clipboard",
    "  cut rich\tDelete selection and copy formatted text to clipboard",
    "  highlight\tToggle highlight decoration on the anchor or the selected text",
    "  italic\tToggle italic decoration on the anchor or the selected text",
    "  strike\tToggle strikethrough decoration on the anchor or the selected text",
    "  underline\tToggle underline decoration on the anchor or the selected text",
    "",
    "\033[1mNOTE ACTIONS – IMAGE MODE\033[0m",
    "  original\tRestore original image size",
    "  aspect ratio\tPreserve aspect ratio while resizing",
    "  scale\tScale the image to preserve its aspect ratio",
    "",
    "\033[1mNOTE ACTIONS – CHECKLIST MODE\033[0m",
    "  save\tSave changes to file",
    "  select all\tCheck all checkbox",
    "  unselect all\tUncheck all checkbox",
]



#!/usr/bin/python3
import re
from pathlib import Path
from typing import Union
from PyQt5 import QtWidgets, QtCore, QtGui

try:
    from QTermWidget import QTermWidget
except ImportError:
    QTermWidget = QtCore.QObject

try:
    from ..backend import UserDirs, RootDirs, logger
    from ..notes import Note
except (ValueError, ImportError):
    from backend import UserDirs, RootDirs, logger
    from notes import Note

log = logger.new(__name__)


class Console(Note):
    mode = "console"

    def __init__(self, core, path: Path):
        super().__init__(core, path)
        self._initNoteWindow(path)
        self.body = Terminal(self)
        self.gridLayout.addWidget(self.body, 1, 0, 1, 3)

    def reload(self):
        """Closes and reload the current terminal"""
        self.body.deleteLater()
        self.body = Terminal(self)
        self.gridLayout.addWidget(self.body, 1, 0, 1, 3)
        self.body.setFocus(True)
        self._zoom()


class Terminal(QTermWidget):
    def __init__(self, note):
        shebang = self._shebang(note.path)

        env = ["TERM=xterm", "LINES=", "COLUMNS="]
        if note.mode == "vim":
            super().__init__(0)
            self.setEnvironment(env)
        elif shebang:  # Script
            super().__init__(0)
            self.setEnvironment(env)
            self.setAutoClose(False)
            self.setShellProgram(shebang)
            self.setArgs([str(note.path)])
            self.startShellProgram()
        else:  # Console
            super().__init__(0)
            self.setEnvironment(env)
            self.startShellProgram()

        self.note = note
        self._setupUi()

    def _exit(self):
        """Cleanly discards parent note when 'exit' is called from terminal"""
        note = self.note
        note.core.loaded[note.path].close()

    def _lostFocus(self):
        """Propagates focus out event to Note"""
        event = QtGui.QFocusEvent(QtCore.QEvent.FocusOut)
        QtCore.QCoreApplication.postEvent(self.note, event)

    def _setupUi(self):
        policy = QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding
        self.setSizePolicy(*policy)
        self.setFocusPolicy(QtCore.Qt.StrongFocus)
        self.setFlowControlEnabled(False)
        self.finished.connect(self._exit)
        self.termLostFocus.connect(self._lostFocus)

        db = self.note.core.sdb["terminal"]
        font = self.font()
        font.setFamily(db["font family"])
        font.setPointSize(db["font size"])
        colorscheme = Path(db["color scheme"]).stem

        hasCustomSchemes = [f for f in UserDirs.TERMINAL.glob("*.colorscheme") if f.is_file()]
        configDir = UserDirs.TERMINAL if hasCustomSchemes else RootDirs.TERMINAL
        self.setTerminalFont(font)
        self.addCustomColorSchemeDir(str(configDir))
        self.setColorScheme(colorscheme)
        self.setTerminalSizeHint(False)

    def _shebang(self, path: Path) -> Union[str, None]:
        """Determines if a file is a script by parsing its shebang (#!)"""
        with open(path, encoding="utf-8") as f:
            r = re.findall(r"^#!(\/\w+.*)", f.read())
        return r[0] if r else None



#!/usr/bin/python3
from PyQt5 import QtGui, QtCore, QtWidgets
from PyQt5.QtWidgets import QLabel

try:
    from ..__id__ import (
        APP_NAME,
        DATE,
        VERSION,
        DESC_AUTHOR,
        DESC_LONG,
        DESC_SHORT,
        LINK_AUR,
        LINK_DONATION,
        LINK_HOMEPAGE,
        LINK_SOURCE,
    )
    from ..settings.base import VSpacer
except (ValueError, ImportError):
    from __id__ import (
        APP_NAME,
        DATE,
        VERSION,
        DESC_AUTHOR,
        DESC_LONG,
        DESC_SHORT,
        LINK_AUR,
        LINK_DONATION,
        LINK_HOMEPAGE,
        LINK_SOURCE,
    )
    from settings.base import VSpacer


class PageAbout(QtWidgets.QWidget):
    def __init__(self):
        super().__init__()
        layout = QtWidgets.QVBoxLayout(self)
        layout.addWidget(About(self))
        layout.addItem(VSpacer())


class About(QtWidgets.QWidget):
    def __init__(self, parent):
        super().__init__(parent)
        font = QtGui.QFont()
        font.setPointSize(32)
        self.titleLabel = QLabel(APP_NAME)
        self.titleLabel.setFont(font)

        font = QtGui.QFont()
        font.setPointSize(14)
        self.versionLabel = QLabel(f"Version {VERSION} — {DATE}")
        self.versionLabel.setFont(font)
        self.copyrightLabel = QLabel(DESC_AUTHOR)

        self.shortDescLabel = QLabel(DESC_SHORT)
        self.shortDescLabel.setWordWrap(True)
        self.longDescLabel = QLabel(DESC_LONG)
        self.longDescLabel.setWordWrap(True)

        self.landingLabel = Hyperlink(f"<a href='{LINK_HOMEPAGE}'>Homepage</a>")
        self.gitLabel = Hyperlink(f"<a href='{LINK_SOURCE}'>Source code</a>")
        self.aurLabel = Hyperlink(f"<a href='{LINK_AUR}'>Arch User Repository</a>")

        font = QtGui.QFont()
        font.setPointSize(12)
        style = "style='color: #A00040; text-decoration:none'"
        self.donateLabel = Hyperlink(
            f"<a href='{LINK_DONATION}' {style}>\u2764\ufe0f Please donate to support development</a>"
        )
        self.donateLabel.setFont(font)

        self.gridLayout = QtWidgets.QGridLayout(self)
        self.gridLayout.addWidget(self.titleLabel, 0, 0, 1, 3)
        self.gridLayout.addWidget(self.versionLabel, 1, 0, 1, 3)
        self.gridLayout.addWidget(self.copyrightLabel, 2, 0, 1, 3)
        self.gridLayout.addItem(VSpacer(height=10), 3, 0, 1, 3)
        self.gridLayout.addWidget(self.shortDescLabel, 4, 0, 1, 3)
        self.gridLayout.addWidget(self.longDescLabel, 5, 0, 1, 3)
        self.gridLayout.addWidget(self.landingLabel, 6, 0, 1, 1)
        self.gridLayout.addWidget(self.gitLabel, 6, 1, 1, 1)
        self.gridLayout.addWidget(self.aurLabel, 6, 2, 1, 1)
        self.gridLayout.addItem(VSpacer(height=25), 7, 2, 1, 1)
        self.gridLayout.addWidget(self.donateLabel, 8, 0, 1, 3)
        self.gridLayout.setContentsMargins(25, 5, 25, 25)
        self.gridLayout.setSpacing(11)


class Hyperlink(QtWidgets.QLabel):
    def __init__(self, label: str):
        super().__init__()
        self.setText(label)
        self.setOpenExternalLinks(True)
        self.setMouseTracking(True)
        self.linkActivated.connect(self.openLink)

    def openLink(self, url: str):
        url = QtCore.QUrl(url)
        QtGui.QDesktopServices.openUrl(url)



#!/usr/bin/python3
import argparse
import shutil
import sys
from pathlib import Path
from typing import Tuple
from PyQt5 import QtCore, QtDBus, QtWidgets

try:
    from .__id__ import ID, HELP
    from .backend import UserDirs, UserFiles, RootDirs, logger
    from .backend.cli import parser
except ImportError:
    from __id__ import ID, HELP
    from backend import UserDirs, UserFiles, RootDirs, logger
    from backend.cli import parser

QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling, True)
QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_UseHighDpiPixmaps, True)
log = logger.new(__name__)


def initialize(args: argparse.Namespace):
    service = f"org.{ID}"
    path = f"/org/{ID}"
    bus = QtDBus.QDBusConnection.sessionBus()
    interface = QtDBus.QDBusInterface(service, path, "", bus)

    if interface.isValid():
        # FIXME: Can a Namespace object be sent through DBus ?
        interface.call("ParseCommands", sys.argv[1:])
        sys.exit(0)
    else:
        with open(UserFiles.LOG, "w", encoding="utf-8"):
            log.info("Cleaned previous session log")

        Setup.install()
        if args.reset:
            Setup.reset(args.reset)
        try:
            app = __import__(f"{ID}.main", fromlist=["main"])
        except ImportError:
            import main as app
        app.main(args)


def main():
    args = parser(sys.argv[1:])  # Early filter for help or malformed commands
    if args.help:
        for h in HELP:
            print(h.expandtabs(18))
        sys.exit(0)
    initialize(args)


class Setup:
    def install():
        """Creates userspace directories and shortcuts"""

        def _dirs():
            dirs = (
                UserDirs.NOTES,
                UserDirs.PALETTES,
                UserDirs.STYLES,
                UserDirs.TERMINAL,
                UserDirs.VIM,
            )
            for path in dirs:
                path.mkdir(parents=True, exist_ok=True)

        def _linksLinux():
            links = {
                UserDirs.PALETTES / "default": RootDirs.PALETTES,
                UserDirs.STYLES / "default": RootDirs.STYLES,
                UserDirs.TERMINAL / "default": RootDirs.TERMINAL,
                UserDirs.VIM / "default": RootDirs.VIM,
            }
            for path, target in links.items():
                try:
                    path.symlink_to(target, target_is_directory=True)
                except FileExistsError:
                    pass

        def _linksWindows():
            links = {
                UserDirs.PALETTES / "default.lnk": RootDirs.PALETTES,
                UserDirs.STYLES / "default.lnk": RootDirs.STYLES,
                UserDirs.TERMINAL / "default.lnk": RootDirs.TERMINAL,
                UserDirs.VIM / "default.lnk": RootDirs.VIM,
            }
            from win32com.client import Dispatch

            shell = Dispatch("WScript.Shell")
            for path, target in links.items():
                path, target = str(path), str(target)
                shortcut = shell.CreateShortCut(path)
                shortcut.Targetpath = target
                shortcut.save()

        def _readme():
            path = UserDirs.UI / "readme.txt"
            if path.exists():
                return

            with open(path, "w", encoding="utf-8") as f:
                f.write(
                    "Users can create their own styles and palettes simply by dropping new CSS\n"
                    "files in the relevant folders. In case of a file name collision, default\n"
                    "configuration is overridden. Default styles and palettes can be ignored\n"
                    "by adding new file names under 'general/ignored css' (settings.json).\n\n"
                    "Vim and QTermWidget palettes are also fully customizable. A typical\n"
                    "approach is to copy a file from the 'default' folders and to edit it.\n"
                    "Qt's CSS documentation and QTermWidget color schemes can be found at:\n\n"
                    "https://doc.qt.io/qt-5/stylesheet-syntax.html\n"
                    "https://doc.qt.io/qt-5/stylesheet-reference.html\n"
                    "https://stuff.mit.edu/afs/athena/software/texmaker_v5.0.2/qt57/doc/qtwidgets/stylesheet-examples.html\n"
                    "https://github.com/lxqt/qtermwidget/tree/master/lib/color-hasCustomSchemes"
                )

        _dirs()
        _readme()
        if sys.platform.startswith("win"):
            _linksWindows()
        else:
            _linksLinux()

    @classmethod
    def reset(cls, args: str):
        """Resets userspace configuration files"""
        if args == "all":
            cls.reset("settings")
            cls.reset("profiles")

        elif args == "settings":
            cls._deleteFiles((UserFiles.SETTINGS, UserFiles.NOTES))

        elif args == "profiles":
            cls._deleteFiles((UserFiles.PROFILES,))

    def _copyFile(src: Path, dest: Path):
        """Copies a file and creates its parent tree"""
        dest.parent.mkdir(parents=True, exist_ok=True)
        shutil.copyfile(src, dest)

    @classmethod
    def _deleteFiles(cls, paths: Tuple[Path]):
        """Removes files without errors"""
        for path in paths:
            try:
                path.unlink()
                log.warning(f"Removed user file '{cls._shortName(path)}'")
            except PermissionError:
                log.error(f"PermissionError: Could not delete '{cls._shortName(path)}'")
            except FileNotFoundError:
                log.error(f"FileNotFoundError: Could not delete '{cls._shortName(path)}'")

    @classmethod
    def _hasChanges(cls, root: Path, user: Path) -> bool:
        """Verifies if two files are the same"""
        with open(root, encoding="utf-8") as f1, open(user, encoding="utf-8") as f2:
            return f1.read() != f2.read()

    @classmethod
    def _overwrite(cls, root: Path, user: Path):
        """Prompts user before overwriting existing configuration"""
        log.warning(f"Resetting '{cls._shortName(user)}'")
        if cls._prompt(f"This will overwrite all changes made in '{user.name}'. Continue?"):
            cls._copyFile(root, user)
            log.warning(f"Overwritten user file '{cls._shortName(user)}'")

    def _prompt(question: str) -> bool:
        """Command-line interface prompt for user confirmation"""
        reply = None
        while reply not in ("", "y", "n"):
            reply = input(f"{question} (Y/n): ").lower()
        return reply in ("", "y")

    @classmethod
    def _resetFiles(cls, files: tuple):
        """Copies non-existent configuration, else prompt user for overwriting"""
        for root, user in files:
            if not user.exists():
                cls._copyFile(root, user)
                log.info(f"Copied default file '{cls._shortName(user)}'")
            elif cls._hasChanges(root, user):
                cls._overwrite(root, user)

    def _shortName(path: Path) -> str:
        """Replaces path to home directory with ~"""
        return str(path).replace(str(Path.home()), "~")


if __name__ == "__main__":
    main()



#!/usr/bin/python3
import logging
import os
import platform
import shutil
import sys
from datetime import datetime

try:
    from ..backend import UserDirs, UserFiles
except (ValueError, ImportError):
    from backend import UserDirs, UserFiles

logging.addLevelName(60, "QUIET")


class ConsoleHandler(logging.StreamHandler):
    VERBOSE = logging.INFO

    def __init__(self):
        super().__init__()
        format_ = "%(asctime)s [%(levelname)s] %(name)s : %(message)s"
        formatter = logging.Formatter(format_, datefmt="%H:%M:%S")
        self.setFormatter(formatter)
        self.setLevel(self.VERBOSE)


class FileHandler(logging.FileHandler):
    def __init__(self):
        super().__init__(filename=UserFiles.LOG)
        formatter = logging.Formatter(
            "%(asctime)s [%(levelname)s] %(name)s, %(funcName)s() : %(message)s",
            datefmt="%d/%m/%Y %H:%M:%S",
        )
        self.setFormatter(formatter)
        self.setLevel(logging.DEBUG)


def new(name: str) -> logging.Logger:
    """Returns a per-module logger instance"""
    UserDirs.LOGS.mkdir(parents=True, exist_ok=True)
    logger = logging.getLogger(name)
    logger.addHandler(ConsoleHandler())
    logger.addHandler(FileHandler())
    logger.setLevel("DEBUG")
    logger.propagate = False
    logger.disable_existing_loggers = False
    return logger


def _exception(type_, value, tb):
    """Global exception handler. Saves crash logs with traceback"""

    def environment() -> str:
        env = [f"{v}={os.environ.get(v)}" for v in ("DESKTOP_SESSION", "QT_QPA_PLATFORMTHEME", "SHELL")]
        log.info(platform.platform() + ", Python " + platform.python_version())
        log.info(", ".join(env))

    try:
        log = new(__name__)
        environment()
        while "tb_next" in dir(tb):  # Log traceback stack
            frame = tb.tb_frame
            code = frame.f_code
            log.critical(f"file '{code.co_filename}' in {code.co_name} (line {frame.f_lineno})")
            tb = tb.tb_next
        log.critical(f"{type_} {value}".rstrip())  # Log exception

        # Save crash log
        now = datetime.now()
        date = f"{now.year:02d}-{now.month:02d}-{now.day:02d}"
        time = f"{now.hour:02d};{now.minute:02d};{now.second:02d}"
        logfile = UserDirs.LOGS / f"crash {date} {time}.log"
        shutil.copy(UserFiles.LOG, logfile)
        print(f'\n    //  \\\\\n   _\\\\()//_   "{logfile}"\n  / //  \\\\ \\\n   | \\__/ |\n')
    except Exception:
        pass
    finally:  # Ensures default hook is called
        sys.__excepthook__(type_, value, tb)
        sys.exit(1)


sys.excepthook = _exception



#!/usr/bin/python3
from collections import namedtuple
from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtWidgets import QLabel

try:
    from ..backend.database import AutoCheckBox, AutoSpinBox
    from ..settings.base import HSpacer, VSpacer, Page
except (ValueError, ImportError):
    from backend.database import AutoCheckBox, AutoSpinBox
    from settings.base import HSpacer, VSpacer, Page

IconsColor = namedtuple("IconsColor", ("tray", "menu"))
ICON_GROUP_TO_KEY = {"Context menus icons": "menu-icon", "Tray icon": "tray-icon"}


class PageGeneral(Page):
    def __init__(self, *args):
        super().__init__(*args)
        self.icons = Icons(self)
        layout = QtWidgets.QVBoxLayout(self)
        layout.addWidget(General(self))
        layout.addItem(HSpacer())
        layout.addWidget(MessageBar(self))
        layout.addItem(HSpacer())
        layout.addWidget(Launcher(self))
        layout.addItem(HSpacer())
        layout.addWidget(self.icons)
        layout.addItem(VSpacer())


class General(QtWidgets.QGroupBox):
    def __init__(self, parent):
        super().__init__("General", parent)
        hotkeysBox = AutoCheckBox("Enable hotkeys", parent.db, ("general", "hotkeys"))
        minBox = AutoCheckBox("Minimize pinned notes on startup", parent.db, ("general", "minimize"))
        skipBox = AutoCheckBox("Skip taskbar", parent.db, ("general", "skip taskbar"))
        repositionBox = AutoCheckBox(
            "Snap out of bounds notes on screen grid", parent.db, ("general", "reposition")
        )
        copyBox = AutoCheckBox(
            "Enable rich text internal copy/paste", parent.db, ("general", "accept qrichtext")
        )
        fileMonitorBox = AutoCheckBox("Automatically open notes created externally", parent.db, ("general", "file monitor"))

        skipBox.setToolTip("Hide the visible notes from system taskbar")
        repositionBox.setToolTip("On screen change, move out of\nbound notes back on screen")
        copyBox.setToolTip("Preserve formatting when copy/pasting\ninto the Rich Text mode")

        elideBox = AutoSpinBox(parent.db, ("general", "elide threshold"))
        elideBox.setToolTip("Title elide threshold. Enter '0' to disable")
        elideFrame = QtWidgets.QWidget()
        elideLayout = QtWidgets.QHBoxLayout(elideFrame)
        elideLayout.setContentsMargins(0, 0, 0, 0)
        elideLayout.addWidget(QtWidgets.QLabel("Title elide threshold"))
        elideLayout.addWidget(elideBox)
        elideLayout.addItem(HSpacer())

        layout = QtWidgets.QVBoxLayout(self)
        layout.addWidget(hotkeysBox)
        layout.addWidget(minBox)
        layout.addWidget(skipBox)
        layout.addWidget(repositionBox)
        layout.addWidget(copyBox)
        layout.addWidget(fileMonitorBox)
        layout.addWidget(elideFrame)


class MessageBar(QtWidgets.QGroupBox):
    def __init__(self, parent):
        super().__init__("Message bar", parent)
        folderBox = AutoCheckBox("Current folder", parent.db, ("message bar", "folder"))
        wordBox = AutoCheckBox("Word count", parent.db, ("message bar", "words"))
        pixelBox = AutoCheckBox("Pixel count", parent.db, ("message bar", "pixels"))
        encryptBox = AutoCheckBox("Encryption icon", parent.db, ("message bar", "encryption"))
        popupBox = AutoSpinBox(parent.db, ("message bar", "popup interval"))
        popupBox.setSingleStep(100)

        folderBox.setToolTip("Display the current folder and subfolders, if any")
        wordBox.setToolTip("Show a word and characters counter (text mode)")
        pixelBox.setToolTip("Show the size in pixels (image mode)")
        encryptBox.setToolTip("Show encryption indicator icon")
        popupBox.setToolTip("Popup message duration (ms)")
        self.setToolTip("The message bar display basic information at the note bottom")

        layout = QtWidgets.QGridLayout(self)
        layout.addWidget(folderBox, 0, 0, 1, 1)
        layout.addWidget(wordBox, 0, 1, 1, 1)
        layout.addWidget(pixelBox, 0, 2, 1, 1)
        layout.addWidget(encryptBox, 1, 0, 1, 1)
        layout.addWidget(QLabel("Popup duration (ms)"), 1, 1, 1, 1)
        layout.addWidget(popupBox, 1, 2, 1, 1)


class Launcher(QtWidgets.QGroupBox):
    def __init__(self, parent):
        super().__init__("Launcher", parent)
        caseBox = AutoCheckBox("Case sensitive", parent.db, ("launcher", "case"))
        hideBox = AutoCheckBox("Hide on launch", parent.db, ("launcher", "hide"))

        hideBox.setToolTip("Close the dialog once a note is chosen")
        self.setToolTip(
            "The launcher is a standalone dialog which allow to quickly search\n"
            "among the note repository. It is accessible through command line\n"
            "interface or 'Search repository' core action"
        )

        layout = QtWidgets.QHBoxLayout(self)
        layout.addWidget(caseBox)
        layout.addWidget(hideBox)


class KeyCombo(QtWidgets.QComboBox):
    def __init__(self, parent):
        super().__init__()
        self.parent = parent
        self.addItems(["Context menus icons", "Tray icon"])

    def currentKey(self) -> str:
        return ICON_GROUP_TO_KEY[self.currentText()]


class Icons(QtWidgets.QGroupBox):
    def __init__(self, parent):
        super().__init__("Global icons color", parent)
        self.parent = parent
        self.color = IconsColor(self.currentColor("tray-icon"), self.currentColor("menu-icon"))
        self.colors = {
            "tray-icon": self.color.tray,
            "menu-icon": self.color.menu,
        }
        darkButton = QtWidgets.QPushButton("Dark")
        lightButton = QtWidgets.QPushButton("Light")
        customButton = QtWidgets.QPushButton("Custom")

        darkButton.setToolTip("Dark icon color for light themes")
        lightButton.setToolTip("Light icon color for dark themes")
        customButton.setToolTip("Choose a custom icon color")
        self.setToolTip("These shortcut allow to quickly change the icon color\nof all context menus.")

        icon = parent.core.icons["toggle"]
        icons = {darkButton: "#404040", lightButton: "#8f8f8f", customButton: "#d4af37"}
        for button, color in icons.items():
            color = QtGui.QColor(color)
            icon = parent.core.colorize(icon, color)
            button.setIcon(icon)
            button.setFocusPolicy(QtCore.Qt.NoFocus)
            button.clicked.connect(self.setColor)
            button.color = color if button != customButton else None

        self.keyCombo = KeyCombo(self)
        layout = QtWidgets.QGridLayout(self)
        layout.addWidget(QLabel("Current group :"), 0, 0, 1, 1)
        layout.addWidget(self.keyCombo, 0, 1, 1, 2)
        layout.addWidget(darkButton, 1, 0, 1, 1)
        layout.addWidget(lightButton, 1, 1, 1, 1)
        layout.addWidget(customButton, 1, 2, 1, 1)

    def colorInputDialog(self) -> QtGui.QColor:
        """Opens a color picker dialog"""
        key = self.keyCombo.currentKey()
        initial = self.colors[key]
        target = self.keyCombo.currentText().lower()
        dialog = QtWidgets.QColorDialog(initial, self.parent)
        dialog.setWindowTitle(f"Pick a color for {target}")
        dialog.exec_()
        color = dialog.selectedColor()
        return color if color.isValid() else initial

    def currentColor(self, key: str) -> QtGui.QColor:
        """Returns the current menu icon color fetched from CSS files"""
        css = self.parent.core.getNoteDecorationsCSS()
        return QtGui.QColor(css.get(key))

    def setColor(self):
        """Schedules the choosen color to be applied once the parent dialog is accepted"""
        color = QtCore.QObject.sender(self).color
        color = color if color else self.colorInputDialog()
        key = self.keyCombo.currentKey()
        self.colors[key] = color



#!/usr/bin/python3
import argparse
from pathlib import Path
from PyQt5 import QtCore, QtDBus

try:
    from ..__id__ import ID
    from ..backend import DeepPath, UserDirs, logger
except (ValueError, ImportError):
    from __id__ import ID
    from backend import DeepPath, UserDirs, logger

log = logger.new(__name__)


def parser(argv: list) -> argparse.Namespace:
    """Setups the command line interface, validates and parses command line options"""
    parser = argparse.ArgumentParser(prog=ID, add_help=False, allow_abbrev=False)
    parser.add_argument("-h", "--help", action="store_true")
    parser.add_argument("-q", "--quit", action="store_true")
    parser.add_argument(
        "-v",
        "--verbose",
        action="store",
        choices=["debug", "info", "warning", "error", "critical", "quiet"],
    )
    parser.add_argument("-c", "--core", action="store", type=str, nargs="+", default=[])
    parser.add_argument("-n", "--note", action="store", type=str, nargs="+")
    parser.add_argument("-p", "--path", action="store", type=str, nargs=1)
    parser.add_argument(
        "-s", "--search", action="store", nargs="?", const="toggle", choices=["toggle", "show", "hide"]
    )
    parser.add_argument(
        "-r",
        "--reset",
        action="store",
        nargs="?",
        const="all",
        choices=["all", "profiles", "settings"],
    )
    args = parser.parse_args(argv)
    if args.note and args.path is None:
        parser.error("--note requires --path option")
    elif args.note and args.path:
        path = Path(args.path[0])
        if (UserDirs.NOTES / path).is_file():
            args.path[0] = UserDirs.NOTES / path
        elif not path.expanduser().is_file():
            parser.error(f"Invalid path '{path}'")

    if args.quit:
        args.core.append("quit")

    if args.verbose:
        from backend import logger

        logger.ConsoleHandler.VERBOSE = args.verbose.upper()
    return args


class QDBusObject(QtCore.QObject):
    def __init__(self, core):
        QtCore.QObject.__init__(self)
        self.__dbusAdaptor = QDBusServerAdapter(self, core)
        self.start()

    def start(self):
        """Initializes D-Bus session"""
        bus = QtDBus.QDBusConnection.sessionBus()
        bus.registerObject(f"/org/{ID}", self)
        bus.registerService(f"org.{ID}")
        return bus


class QDBusServerAdapter(QtDBus.QDBusAbstractAdaptor):
    QtCore.Q_CLASSINFO("D-Bus Interface", f"org.{ID}")
    QtCore.Q_CLASSINFO(
        "D-Bus Introspection",
        f'<interface name="org.{ID}">\n'
        '  <method name="ParseCommands">\n'
        '    <arg direction="in" type="a{s}" name="commands"/>\n'
        "  </method>\n"
        "</interface>\n",
    )

    def __init__(self, server, core):
        super().__init__(server)
        self.core = core

    @QtCore.pyqtSlot(list)
    def ParseCommands(self, argv: list):
        """Forwards the parsed arguments Namespace to the running instance"""
        args = QtCore.Q_ARG(object, parser(argv))
        QtCore.QMetaObject.invokeMethod(self.core.parser, "fromNamespace", QtCore.Qt.QueuedConnection, args)


class CommandParser(QtCore.QObject):
    def __init__(self, core):
        super().__init__()
        self.core = core

    def fromDict(self, args: dict):
        """Parses internal commands"""
        if "core" in args:  # cmd = {'core': ['new']}
            self._core(key for key in args["core"])

        if "note" in args:  # cmd = {'note': ['rename', 'Untitled 1.txt']}
            action, path = args["note"]
            path = DeepPath(path).expanduser()
            self._note([action], path)

    @QtCore.pyqtSlot(argparse.Namespace)
    def fromNamespace(self, args: argparse.Namespace):
        """Parses external commands received through the command line interface (case insensitive)"""
        if args.core:
            actions = [key.lower() for key in args.core]
            self._core(actions)

        if args.note:
            actions = [key.lower() for key in args.note]
            path = DeepPath(args.path[0]).expanduser()
            self._note(actions, path)

        if args.search:
            options = {"show": True, "hide": False, "toggle": None}
            self.core.searchToggle(options[args.search])

    def _core(self, cmd: list):
        """Handler for core actions"""
        for action in cmd:
            if action in self.core.actions.tray:
                action = self.core.actions.tray[action]
                log.info(f"Core : {action.label}")
                action.call()
            else:
                log.error(f"Invalid action '{action}'")

    def _note(self, cmd: list, path: DeepPath):
        """Handler for note actions"""
        if path not in self.core.loaded:
            self.core.notes.add(path)
        note = self.core.loaded[path]

        for action in cmd:
            try:
                action = note.actions[action]
                log.info(f"{note.id} : {action.label}")
                try:
                    action.call()
                except TypeError:
                    action.call(path)
                note.popupFrame.showMessage(action.label, "")
            except KeyError:
                log.error(f"Invalid action '{action}'")



#!/usr/bin/python3
from copy import deepcopy
from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtWidgets import QSizePolicy
from PyQt5.QtWidgets import QDialogButtonBox as QButtonBox

try:
    from ..__id__ import APP_NAME, ID
    from ..backend import RootDirs, logger
    from ..settings.general import PageGeneral
    from ..settings.profile import PageProfile
    from ..settings.mouse import PageMouse
    from ..settings.hotkeys import PageHotkeys
    from ..settings.menus import PageCoreMenu, PageNoteContextMenu, PageNoteToolbarMenu
    from ..settings.logs import PageLogs
    from ..settings.advanced import PageAdvanced
    from ..settings.plugins import PagePlugins
    from ..settings.about import PageAbout
    from ..settings.reset import ResetDialog
except (ValueError, ImportError):
    from __id__ import APP_NAME, ID
    from backend import RootDirs, logger
    from settings.general import PageGeneral
    from settings.profile import PageProfile
    from settings.mouse import PageMouse
    from settings.hotkeys import PageHotkeys
    from settings.menus import PageCoreMenu, PageNoteContextMenu, PageNoteToolbarMenu
    from settings.logs import PageLogs
    from settings.advanced import PageAdvanced
    from settings.plugins import PagePlugins
    from settings.about import PageAbout
    from settings.reset import ResetDialog

log = logger.new(__name__)


class Settings(QtWidgets.QDialog):
    def __init__(self, core):
        super().__init__()
        self.core = core
        self.setWindowTitle(f"{APP_NAME} Settings")
        self.resize(580, 500)

        icon = QtGui.QIcon(str(RootDirs.ICONS / f"{ID}_settings.svg"))
        self.setWindowIcon(icon)

        items = [
            "General",
            "Profile default",
            "Mouse events",
            "Hotkeys",
            "Core menus",
            "Note context menus",
            "Note toolbars menus",
            "Logs and archives",
            "Advanced",
            "Plugins status",
            "About",
        ]
        self.sideMenu = QtWidgets.QListWidget()
        self.sideMenu.setFocusPolicy(QtCore.Qt.StrongFocus)
        self.sideMenu.addItems(items)
        self.sideMenu.item(0).setSelected(True)
        self.sideMenu.selectionModel().selectionChanged.connect(self.setPageIndex)
        fm = QtGui.QFontMetrics(self.sideMenu.font())
        width = int(fm.horizontalAdvance(max(items)) * 1.8)
        self.sideMenu.setMinimumWidth(width)

        self.dbCopy = deepcopy(dict(core.sdb))
        self.general = PageGeneral(core, self.dbCopy)
        self.stack = QtWidgets.QStackedWidget()
        self.stack.addWidget(self.general)
        self.stack.addWidget(PageProfile(core, self.dbCopy))
        self.stack.addWidget(PageMouse(core, self.dbCopy))
        self.stack.addWidget(PageHotkeys(core, self.dbCopy))
        self.stack.addWidget(PageCoreMenu(core, self.dbCopy))
        self.stack.addWidget(PageNoteContextMenu(core, self.dbCopy))
        self.stack.addWidget(PageNoteToolbarMenu(core, self.dbCopy))
        self.stack.addWidget(PageLogs(core, self.dbCopy))
        self.stack.addWidget(PageAdvanced(core, self.dbCopy))
        self.stack.addWidget(PagePlugins())
        self.stack.addWidget(PageAbout())

        splitter = QtWidgets.QSplitter()
        splitter.setOrientation(QtCore.Qt.Horizontal)
        splitter.addWidget(self.sideMenu)
        splitter.addWidget(self.stack)
        layout = QtWidgets.QVBoxLayout(self)
        layout.addWidget(splitter)
        layout.addWidget(self.BottomFrame(self))
        self.show()

    def accept(self):
        super().accept()
        self.apply()

    def apply(self):
        """Applies new application settings"""
        for key, color in self.general.icons.colors.items():
            new = color.name()
            old = self.general.icons.currentColor(key).name()
            if new != old:
                self.core.setGlobalIconsColor(key, new)

        self.core.sdb.update(self.dbCopy)
        self.core.fileMonitor.resume()
        log.info("New settings applied")

    def reset(self):
        """Handles restore to default dialog"""
        dialog = ResetDialog(self.core)
        if dialog.exec() and dialog.isChecked():
            self.accept()
            for key, item in dialog.checkboxes.items():
                if item.isChecked():
                    dialog.resetKey(key)
            self.core.settings()

    def setPageIndex(self):
        """Updates the stack widget according to side menu changes"""
        index = self.sideMenu.currentRow()
        self.stack.setCurrentIndex(index)

    class BottomFrame(QtWidgets.QWidget):
        def __init__(self, parent):
            super().__init__()
            self.parent = parent
            self.setSizePolicy(QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Maximum))

            buttons = QButtonBox.Cancel | QButtonBox.Ok | QButtonBox.Apply | QButtonBox.RestoreDefaults
            self.buttonBox = QButtonBox()
            self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
            self.buttonBox.setStandardButtons(buttons)
            self.buttonBox.accepted.connect(parent.accept)
            self.buttonBox.rejected.connect(parent.reject)
            self.buttonBox.clicked.connect(self._clicked)

            layout = QtWidgets.QHBoxLayout(self)
            layout.setContentsMargins(0, 0, 0, 0)
            layout.addWidget(self.buttonBox)

        def _clicked(self, button: QtWidgets.QPushButton):
            """Parses clicked button type"""
            button = self.buttonBox.standardButton(button)
            if button == QButtonBox.Apply:
                self.parent.apply()
            elif button == QButtonBox.RestoreDefaults:
                self.parent.reset()



#!/usr/bin/python3
import hashlib
import json
import re
import shutil
from pathlib import Path

try:
    from ..__id__ import APP_NAME
    from ..__db__ import SDB_DEFAULT
    from ..backend import UserDirs, UserFiles, logger
    from ..menus.core import ConfirmPrompt
except (ValueError, ImportError):
    from __id__ import APP_NAME
    from __db__ import SDB_DEFAULT
    from backend import UserDirs, UserFiles, logger
    from menus.core import ConfirmPrompt

log = logger.new(__name__)


class Legacy:
    @classmethod
    def css(cls, core):
        """Removes stale configuraton files installed prior to version 1.3"""
        if core.sdb["legacy"] < 1.3:
            stales = {
                UserFiles.CSS: "5e36002b1a0fce95a47df9b603335d65",
                UserDirs.PALETTES / "graphite.css": "7ff06b4333796b1c8f5af023052d1991",
                UserDirs.PALETTES / "classic.css": "697c22cf0db5b2c1c77cc09c5a93a7b5",
                UserDirs.PALETTES / "northern.css": "38d66b9c288928fa6f904e7dba06822b",
                UserDirs.STYLES / "toolbar.css": "34515e90227deb2dcc6a34a5775e014a",
                UserDirs.STYLES / "minimal.css": "a27d50e188bbe3eb19e337b6113b3883",
                UserDirs.VIM / "vimrc": "dc0f2ef31112f91093d7ccf1c912b70a",
                UserDirs.VIM / "colors" / "flattown.vim": "ceac229594f9541bce9bb06041c17c39",
                UserDirs.TERMINAL / "default.colorscheme": "dde0fac8fb0cd36750e7fdf599b35658",
            }
            for file, md5 in stales.items():
                if cls._fileEqualMD5(file, md5):
                    file.unlink()
                    log.warning(f"Removed deprecated file '{file}'")

            if UserFiles.CSS.is_file():
                name = UserFiles.CSS.parent / f"{UserFiles.CSS.name}.bak"
                UserFiles.CSS.rename(name)
                log.warning(f"Disabled file '{UserFiles.CSS}'")

            colors = UserDirs.VIM / "colors"
            if colors.is_dir() and colors.is_empty():
                colors.rmdir()

    @classmethod
    def menus(cls, core):
        """Replaces non-customized menus to default prior to version 1.3"""
        if core.sdb["legacy"] < 1.3:
            menus = {
                "context menus": "d91125298788cccfb5cfe4b6e80022a6",
                "toolbar menus": "66430991463994a90955af877474e5a0",
            }
            for key, md5 in menus.items():
                currentMenu = str(core.sdb[key])
                if md5 == cls._md5(currentMenu):
                    core.sdb[key].clear()
                    core.sdb[key].update(SDB_DEFAULT[key])
                    log.warning(f"Replaced default '{key}' content")

    @classmethod
    def profiles(cls, core):
        """Adds 'zoom' property in every profiles"""
        if core.sdb["legacy"] < 1.4:
            for nid, profile in core.pdb.items():
                if "zoom" not in profile:
                    profile["zoom"] = 0

    @classmethod
    def prompt(cls, core):
        """Proposes notes migration from a legacy version (former QtPad users)"""
        qtpad = UserDirs.CONFIG.parent / "qtpad" / "notes"
        if qtpad.is_dir():
            msg = (
                "Notes from an older version were found in the folder\n"
                f"'{qtpad}'\n\n"
                f"Would you like to copy those into {APP_NAME} notes repository ?"
            )
            dialog = ConfirmPrompt("QtPad Notes Migration", msg)
            if dialog.exec_() == dialog.Yes:
                cls._migrate(qtpad)

    @staticmethod
    def renameAntidote(core):
        """Renames antidote database keys: 'antidote' -> 'antidote corrector'"""
        if core.sdb["legacy"] < 1.5:
            sdb = json.dumps(core.sdb)
            sdb = re.sub('"antidote"', '"antidote corrector"', sdb)
            core.sdb.update(json.loads(sdb))

    @staticmethod
    def renameSaveAs(core):
        """Renames settings database keys: 'save as' -> 'export'"""
        if core.sdb["legacy"] < 1.5:
            sdb = json.dumps(core.sdb)
            sdb = re.sub('"save as dir"', '"export directory"', sdb)
            sdb = re.sub('"save as"', '"export"', sdb)
            core.sdb.update(json.loads(sdb))

    @classmethod
    def _fileEqualMD5(cls, file: Path, md5: str) -> bool:
        """Compares a MD5 hash with a MD5 sum of a file"""
        if not file.is_file():
            return False

        with open(file, encoding="utf-8") as f:
            if md5 == cls._md5(f.read()):
                return True
        return False

    def _md5(text: str) -> str:
        """Computes a MD5 hash from a string"""
        h = hashlib.md5(text.encode("utf8"))
        return h.hexdigest()

    def _migrate(src: Path):
        """Copy notes from a legacy version (former QtPad users)"""
        files = [f for f in src.rglob("*.txt") if f.is_file()]
        files += [f for f in src.rglob("*.png") if f.is_file()]
        for f in files:
            dest = UserDirs.NOTES / f.relative_to(src)
            if dest.exists():
                log.warning(f"Ignored migration of '{f}'")
            else:
                dest.parent.mkdir(parents=True, exist_ok=True)
                shutil.copyfile(f, dest)
                log.info(f"Copied '{f}'")



#!/usr/bin/python3
import pythoncom
import win32api
import win32com.client
import win32com.server.util
import win32con
import win32gui
from typing import Callable, Union
from win32com.server import localserver
from PyQt5 import QtCore
from PyQt5.QtGui import QTextCursor
from PyQt5.QtWidgets import QPlainTextEdit

try:
    PACKAGED = True
    from ...__id__ import ID
    from ...backend import logger
except (ValueError, ImportError):
    PACKAGED = False
    from __id__ import ID
    from backend import logger

log = logger.new(__name__)


def getRegKey(hkey: str, path: str, name: str) -> Union[str, int]:
    """Returns value from a registry key"""
    hkey = {"HKLM": win32con.HKEY_LOCAL_MACHINE, "HKCU": win32con.HKEY_CURRENT_USER}[hkey.upper()]
    key = win32api.RegOpenKey(hkey, path)
    value = win32api.RegQueryValueEx(key, name)[0]
    return value


def setRegKey(hkey: str, path: str, value: str):
    """Sets a registry key value"""
    hkey = {"HKLM": win32con.HKEY_LOCAL_MACHINE, "HKCU": win32con.HKEY_CURRENT_USER}[hkey.upper()]
    win32api.RegSetValue(hkey, path, win32con.REG_SZ, value)


class Adapter:
    _reg_clsctx_ = pythoncom.CLSCTX_LOCAL_SERVER
    _public_attrs_ = ["parent"]
    _public_methods_ = [
        "setParent",
        "ActiveApplication",
        "ActiveDocument",
        "DonneDebutSelection",
        "DonneFinSelection",
        "DonneIdDocumentCourant",
        "DonneIdZoneDeTexte",
        "DonneIdZoneDeTexteCourante",
        "DonneIntervalle",
        "DonneLongueurZoneDeTexte",
        "DonneNbZonesDeTexte",
        "DonnePolice",
        "DonneTitreDocCourant",
        "RemplaceIntervalle",
        "SelectionneIntervalle",
    ]

    def setParent(self, parent: object):  # parent: PyIDispatch
        """Converts the PyIDispatch object into an <unknown> COMObject"""
        self.parent = win32com.client.Dispatch(parent)

    @classmethod
    def register(cls, progid: str):
        """Registers a local COM server into the registry"""
        import win32com.server.register

        clsid = str(pythoncom.CreateGuid())
        python = win32com.server.register._find_localserver_exe(1)
        python = win32api.GetShortPathName(python)
        server32 = win32com.server.register._find_localserver_module()
        server32 = f'{python} "{server32}" {clsid}'
        namespace = f"{ID}." if PACKAGED else ""
        namespace = f"{namespace}plugins.antidote.com.{cls.__name__}"

        try:
            setRegKey("HKCU", rf"Software\Classes\{progid}\CLSID", clsid)
            setRegKey("HKCU", rf"Software\Classes\AppID\{clsid}", progid)
            setRegKey("HKCU", rf"Software\Classes\CLSID\{clsid}\LocalServer32", server32)
            setRegKey("HKCU", rf"Software\Classes\CLSID\{clsid}\ProgID", progid)
            setRegKey("HKCU", rf"Software\Classes\CLSID\{clsid}\PythonCOM", namespace)
            log.info(f"Registered server {progid} as {clsid}")
        except BaseException:
            log.error(f"Failed to register COM server {progid}")

    @classmethod
    def unregister(cls, progid: str):
        """Removes the local COM server from the registry"""
        try:
            clsid = getRegKey("HKCU", rf"Software\Classes\{progid}\CLSID", "")
            win32api.RegDeleteTree(win32con.HKEY_CURRENT_USER, rf"Software\Classes\{progid}")
            win32api.RegDeleteTree(win32con.HKEY_CURRENT_USER, rf"Software\Classes\AppID\{clsid}")
            win32api.RegDeleteTree(win32con.HKEY_CURRENT_USER, rf"Software\Classes\CLSID\{clsid}")
            log.info(f"Unregistered COM server {progid}")
        except BaseException:
            log.error(f"Failed to unregister COM server {progid}")

    def ActiveApplication(self):
        self.parent.activeApplication()

    def ActiveDocument(self, idDocument: int):
        self.parent.activeDocument(idDocument)

    def DonneDebutSelection(self, idDocument: int, idZoneDeTexte: int) -> int:
        return self.parent.donneDebutSelection(idDocument, idZoneDeTexte)

    def DonneFinSelection(self, idDocument: int, idZoneDeTexte: int) -> int:
        return self.parent.donneFinSelection(idDocument, idZoneDeTexte)

    def DonneIdDocumentCourant(self) -> int:
        return self.parent.donneIdDocumentCourant()

    def DonneIdZoneDeTexte(self, idDocument: int, indiceZoneDeTexte: int) -> int:
        return self.parent.donneIdZoneDeTexte(idDocument, indiceZoneDeTexte)

    def DonneIdZoneDeTexteCourante(self, idDocument: int) -> int:
        return self.parent.donneIdZoneDeTexteCourante(idDocument)

    def DonneIntervalle(self, idDocument: int, idZoneDeTexte: int, debut: int, fin: int) -> str:
        return self.parent.donneIntervalle(idDocument, idZoneDeTexte, debut, fin)

    def DonneLongueurZoneDeTexte(self, idDocument: int, idZoneDeTexte: int) -> int:
        return self.parent.donneLongueurZoneDeTexte(idDocument, idZoneDeTexte)

    def DonneNbZonesDeTexte(self, idDocument: int) -> int:
        return self.parent.donneNbZonesDeTexte(idDocument)

    def DonnePolice(self, idDocument: int, idZoneDeTexte: int) -> str:
        return self.parent.donnePolice(idDocument, idZoneDeTexte)

    def DonneTitreDocCourant(self) -> str:
        return self.parent.donneTitreDocCourant()

    def RemplaceIntervalle(self, idDocument: int, idZoneDeTexte: int, debut: int, fin: int, chaine: str):
        self.parent.remplaceIntervalle(idDocument, idZoneDeTexte, debut, fin, chaine)

    def SelectionneIntervalle(self, idDocument: int, idZoneDeTexte: int, debut: int, fin: int):
        self.parent.selectionneIntervalle(idDocument, idZoneDeTexte, debut, fin)


class API:  # API Antidote v2.0 r2
    _public_methods_ = [
        "activeApplication",
        "activeDocument",
        "donneDebutSelection",
        "donneFinSelection",
        "donneIdDocumentCourant",
        "donneIdZoneDeTexte",
        "donneIdZoneDeTexteCourante",
        "donneIntervalle",
        "donneLongueurZoneDeTexte",
        "donneNbZonesDeTexte",
        "donnePolice",
        "donneTitreDocCourant",
        "remplaceIntervalle",
        "selectionneIntervalle",
    ]

    def ignoreRuntimeError(func: Callable) -> Callable:
        """Logs API calls and ignores RuntimeError during execution"""

        def wrapper(self, *args) -> any:
            try:
                _return = func(self, *args)
                log.debug(f"{func.__qualname__}{args} -> {repr(_return)}")
                return _return
            except RuntimeError:
                log.debug(f"RuntimeError in {func.__qualname__}{args}")

        return wrapper

    @ignoreRuntimeError
    def activeApplication(self):
        self.body.window().activateWindow()
        self.body.window().raise_()

    def activeDocument(self, idDocument: int):
        self.activeApplication()

    @ignoreRuntimeError
    def donneDebutSelection(self, idDocument: int, idZoneDeTexte: int) -> int:
        ancre = self.body.textCursor().anchor()
        position = self.body.textCursor().position()
        return min(ancre, position)

    @ignoreRuntimeError
    def donneFinSelection(self, idDocument: int, idZoneDeTexte: int) -> int:
        ancre = self.body.textCursor().anchor()
        position = self.body.textCursor().position()
        return max(ancre, position)

    def donneIdDocumentCourant(self) -> int:
        return id(self)

    def donneIdZoneDeTexte(self, idDocument: int, indiceZoneDeTexte: int) -> int:
        return 1

    def donneIdZoneDeTexteCourante(self, idDocument: int) -> int:
        return 1

    @ignoreRuntimeError
    def donneIntervalle(self, idDocument: int, idZoneDeTexte: int, debut: int, fin: int) -> str:
        textCursor = self.body.textCursor()
        textCursor.movePosition(QTextCursor.Start)
        textCursor.movePosition(QTextCursor.NextCharacter, QTextCursor.MoveAnchor, debut)
        textCursor.movePosition(QTextCursor.NextCharacter, QTextCursor.KeepAnchor, fin - debut)
        return textCursor.selectedText()

    @ignoreRuntimeError
    def donneLongueurZoneDeTexte(self, idDocument: int, idZoneDeTexte: int) -> int:
        return len(self.body.toPlainText())

    def donneNbZonesDeTexte(self, idDocument: int) -> int:
        return 1

    def donnePolice(self, idDocument: int, idZoneDeTexte: int) -> str:
        return ""

    @ignoreRuntimeError
    def donneTitreDocCourant(self) -> str:
        return self.body.window().windowTitle()

    @ignoreRuntimeError
    def remplaceIntervalle(self, idDocument: int, idZoneDeTexte: int, debut: int, fin: int, chaine: str):
        if debut >= 0 and fin >= 0:
            textCursor = self.body.textCursor()
            textCursor.setPosition(debut)
            textCursor.movePosition(QTextCursor.NextCharacter, QTextCursor.KeepAnchor, fin - debut)
            textCursor.insertText(chaine)
        self.body.ensureCursorVisible()

    @ignoreRuntimeError
    def selectionneIntervalle(self, idDocument: int, idZoneDeTexte: int, debut: int, fin: int):
        textCursor = self.body.textCursor()
        textCursor.setPosition(debut)
        textCursor.movePosition(QTextCursor.NextCharacter, QTextCursor.KeepAnchor, fin - debut)
        self.body.setTextCursor(textCursor)


class ServerThread(QtCore.QObject):
    def __init__(self):
        super().__init__()

    def run(self):
        pythoncom.CoInitialize()
        localserver.serve([self.clsid])


class Antidote(QtCore.QObject, API):
    ready = QtCore.pyqtSignal()

    def __init__(self, uid: str, path: str, body: QPlainTextEdit):
        super().__init__()
        self.progid = f"{uid}.Antidote"
        self.body = body
        self.api = None

        version = getRegKey("HKLM", r"Software\Druide informatique inc.\Antidote", "VersionAPI")
        version = min([2.0, float(version)])
        self.version = str(version)
        self._launch(path)

    def init(self):
        """Initializes COM interface"""
        self.worker = ServerThread()
        self.workerThread = QtCore.QThread()
        self.workerThread.started.connect(self.worker.run)
        self.workerThread.started.connect(self._setup)
        self.worker.moveToThread(self.workerThread)
        self._serve()

    def corrector(self):
        """Launches Antidote corrector"""
        self._tool("C")

    def dictionnary(self, tool: str = "D_Defaut"):
        """Launches Antidote dictionnary"""
        options = (
            "D",
            "D_Defnitions",
            "D_Synonymes",
            "D_Antonymes",
            "D_Cooccurrences",
            "D_ChampLexical",
            "D_Conjugaison",
            "D_Famille",
            "D_Citations",
            "D_Historique",
            "D_Illustrations",
            "D_Wikipedia",
        )
        tool = tool if tool in options else "D_Defaut"
        self._tool(tool)

    def guide(self, tool: str = "G_Defaut"):
        """Launches Antidote guide"""
        options = (
            "G",
            "G_Orthographe",
            "G_Lexique",
            "G_Grammaire",
            "G_Syntaxe",
            "G_Ponctuation",
            "G_Style",
            "G_Redaction",
            "G_Typographie",
            "G_Phonetique",
            "G_Historique",
            "G_PointsDeLangue",
        )
        tool = tool if tool in options else "G_Defaut"
        self._tool(tool)

    def _launch(self, path: str):
        """Opens Antidote binary"""
        if not win32gui.FindWindow("AntQ", None):
            QtCore.QProcess.startDetached(path, ["-activex"])
            log.info(f"{path} -activex")

    def _serve(self, failed=False):
        """Initializes the local COM server"""
        try:
            clsid = getRegKey("HKCU", rf"Software\Classes\{self.progid}\CLSID", "")
            self.worker.clsid = clsid
            self.workerThread.start()
        except BaseException:
            if failed:
                log.error("Failed to initialize local COM server")
                Adapter.unregister(self.progid)
            else:
                log.info("Local COM server not found in registry")
                Adapter.register(self.progid)
                self._serve(failed=True)

    def _setup(self):
        """Initializes the local COM server along with Antidote API OLE public server"""
        if win32gui.FindWindow("AntQ", None):
            try:
                self.server = win32com.client.Dispatch(f"{self.progid}")
                self.server.setParent(win32com.server.util.wrap(self))
                self.api = win32com.client.Dispatch("Antidote.ApiOle")
                self.ready.emit()
            except BaseException:
                log.error("Server dispatch failed")
                Adapter.unregister(self.progid)
        else:
            QtCore.QTimer.singleShot(1000, self._setup)

    def _tool(self, tool):
        """Launches a specific tool (corrector/dictionnary/guide)"""
        if self.api:
            self.api.LanceOutilDispatch2(self.server, tool, "", self.version)
        else:
            log.error("Could not call Antidote: Local COM server is not ready")



import platform
import shutil
import sys
from pathlib import Path
from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QSizePolicy, QSpacerItem

try:
    from ..__id__ import APP_NAME, ID
    from ..backend import RootDirs, logger
    from ..backend.legacy import Legacy
    from ..menus.core import ConfirmPrompt
except (ValueError, ImportError):
    from __id__ import APP_NAME, ID
    from backend import RootDirs, logger
    from backend.legacy import Legacy
    from menus.core import ConfirmPrompt

log = logger.new(__name__)


class HSpacer(QSpacerItem):
    def __init__(self, height: int, horizontalPolicy: QSizePolicy):
        policy = QSizePolicy.Minimum, horizontalPolicy
        super().__init__(0, height, *policy)


class Label(QtWidgets.QLabel):
    def __init__(self, pointSize: int = 8, text: str = "", bold=False, wrap=False, center=False):
        super().__init__()
        font = QtGui.QFont()
        font.setPointSize(pointSize)
        font.setBold(bold)
        self.setFont(font)
        self.setWordWrap(wrap)
        self.setText(text)
        if center:
            self.setAlignment(Qt.AlignCenter)


class CheckableButton(QtWidgets.QPushButton):
    def __init__(self, path: Path):
        super().__init__()
        sizePolicy = QSizePolicy(QSizePolicy.Maximum, QSizePolicy.Minimum)
        icon = QtGui.QIcon(str(path))
        iconSize = icon.actualSize(QtCore.QSize(1024, 1024))
        self.setStyleSheet("QPushButton:checked {background: #e98d34; border: solid 1px;}")
        self.setSizePolicy(sizePolicy)
        self.setIcon(icon)
        self.setIconSize(iconSize)
        self.setFocusPolicy(Qt.NoFocus)
        self.setAutoExclusive(True)
        self.setCheckable(True)
        self.setFlat(True)
        self.setText("")


class TrayButton(QtWidgets.QPushButton):
    def __init__(self):
        super().__init__("Manage taskbar")
        self.clicked.connect(self._clicked)

    def _clicked(self):
        """Opens MS Windows Taskbar Settings"""
        w7 = platform.release() == "7"
        cmd = "05D7B0F4-2121-4EFF-BF6B-ED3F69B894D9" if w7 else "0DF44EAA-FF21-4412-828E-260A8728E7F1"
        process = QtCore.QProcess()
        process.setArguments(["shell:::{" + cmd + "}"])
        process.setProgram("explorer")
        process.startDetached()


class Page(QtWidgets.QWizardPage):
    def __init__(self, core):
        super().__init__()
        self.core = core
        self.setTitle(f"{APP_NAME} Customization Wizard")

    @property
    def sdb(self) -> dict:
        return self.core.sdb

    def done(self):
        pass


class PageHome(Page):
    def __init__(self, core):
        super().__init__(core)
        self.setTitle("")
        self.setSubTitle("")
        titleLabel = Label(28, wrap=True, center=True, text=f"Welcome to {APP_NAME} !")
        greetLabel = Label(12, wrap=True)
        greetLabel.setText(
            f"This short wizard will guide you through the steps required to customize {APP_NAME} interface so it best fit your needs."
        )
        stepsLabel = Label(12, wrap=True)
        stepsLabel.setText(
            "The following steps will let you choose some default settings :\n\n"
            "‣ Note color palette\n"
            "‣ Menu icon color\n"
            "‣ Tray icon color\n"
            "‣ Note frame style\n"
            "‣ Text editing mode"
        )

        layout = QtWidgets.QVBoxLayout()
        layout.addItem(HSpacer(20, QSizePolicy.Fixed))
        layout.addWidget(titleLabel)
        layout.addItem(HSpacer(30, QSizePolicy.Fixed))
        layout.addWidget(greetLabel)
        layout.addItem(HSpacer(30, QSizePolicy.Fixed))
        layout.addWidget(stepsLabel)
        layout.setContentsMargins(40, 30, 40, 30)
        self.setLayout(layout)


class PagePalette(Page):
    def __init__(self, core):
        super().__init__(core)
        self.setSubTitle("Select a color palette")

        self.frozenButton = CheckableButton(RootDirs.WIZARD / "palette_frozen.png")
        self.graphiteButton = CheckableButton(RootDirs.WIZARD / "palette_graphite.png")
        self.lavenderButton = CheckableButton(RootDirs.WIZARD / "palette_lavender.png")
        self.seasideButton = CheckableButton(RootDirs.WIZARD / "palette_seaside.png")
        self.solarizedButton = CheckableButton(RootDirs.WIZARD / "palette_solarized.png")
        self.watermelonButton = CheckableButton(RootDirs.WIZARD / "palette_watermelon.png")

        layout = QtWidgets.QGridLayout()
        layout.addItem(QSpacerItem(0, 0, QSizePolicy.Minimum, QSizePolicy.Expanding), 0, 0, 1, 1)
        layout.addWidget(self.frozenButton, 1, 0, 1, 1, alignment=Qt.AlignCenter)
        layout.addWidget(self.graphiteButton, 1, 1, 1, 1, alignment=Qt.AlignCenter)
        layout.addWidget(self.lavenderButton, 1, 2, 1, 1, alignment=Qt.AlignCenter)
        layout.addWidget(self.seasideButton, 2, 0, 1, 1, alignment=Qt.AlignCenter)
        layout.addWidget(self.solarizedButton, 2, 1, 1, 1, alignment=Qt.AlignCenter)
        layout.addWidget(self.watermelonButton, 2, 2, 1, 1, alignment=Qt.AlignCenter)
        layout.addItem(QSpacerItem(0, 0, QSizePolicy.Minimum, QSizePolicy.Expanding), 3, 0, 1, 1)
        self.setLayout(layout)

    def done(self):
        """Applies the checked value"""
        if self.frozenButton.isChecked():
            self.sdb["profile default"]["palette"] = "frozen.css"
        elif self.graphiteButton.isChecked():
            self.sdb["profile default"]["palette"] = "graphite.css"
        elif self.lavenderButton.isChecked():
            self.sdb["profile default"]["palette"] = "lavender.css"
        elif self.seasideButton.isChecked():
            self.sdb["profile default"]["palette"] = "seaside.css"
        elif self.solarizedButton.isChecked():
            self.sdb["profile default"]["palette"] = "solarized.css"
        elif self.watermelonButton.isChecked():
            self.sdb["profile default"]["palette"] = "watermelon.css"


class PageMenuColor(Page):
    def __init__(self, core):
        super().__init__(core)
        self.setSubTitle("Select the menu icon color that best match your theme")

        darkLabel = Label(center=True, text="\n<b>Dark menu icons</b> for light themes")
        lightLabel = Label(center=True, text="\n<b>Light menu icons</b> for dark themes")

        self.darkButton = CheckableButton(RootDirs.WIZARD / "icons_dark.png")
        self.lightButton = CheckableButton(RootDirs.WIZARD / "icons_light.png")

        layout = QtWidgets.QGridLayout()
        layout.addItem(QSpacerItem(0, 0, QSizePolicy.Minimum, QSizePolicy.Expanding), 0, 0, 1, 1)
        layout.addWidget(self.darkButton, 1, 0, 1, 1, alignment=Qt.AlignCenter)
        layout.addWidget(self.lightButton, 1, 1, 1, 1, alignment=Qt.AlignCenter)
        layout.addWidget(darkLabel, 2, 0, 1, 1)
        layout.addWidget(lightLabel, 2, 1, 1, 1)
        layout.addItem(QSpacerItem(0, 0, QSizePolicy.Minimum, QSizePolicy.Expanding), 3, 0, 1, 1)
        self.setLayout(layout)

    def done(self):
        if self.darkButton.isChecked():
            self.core.setGlobalIconsColor("menu-icon", "#404040")
        elif self.lightButton.isChecked():
            self.core.setGlobalIconsColor("menu-icon", "#8f8f8f")


class PageTrayColor(Page):
    def __init__(self, core):
        super().__init__(core)
        self.setSubTitle("Select the tray icon color that best match your theme")

        darkLabel = Label(center=True, text="\n<b>Dark tray icon</b> for light themes")
        lightLabel = Label(center=True, text="\n<b>Light tray icon</b> for dark themes")

        sizePolicy = QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)
        self.darkButton = CheckableButton(RootDirs.WIZARD / "tray_dark.png")
        self.lightButton = CheckableButton(RootDirs.WIZARD / "tray_light.png")
        self.darkButton.setSizePolicy(sizePolicy)
        self.lightButton.setSizePolicy(sizePolicy)

        layout = QtWidgets.QVBoxLayout()
        layout.addItem(QSpacerItem(0, 40, QSizePolicy.Minimum, QSizePolicy.Fixed))
        layout.addWidget(lightLabel)
        layout.addWidget(self.lightButton, alignment=Qt.AlignCenter)
        layout.addItem(QSpacerItem(0, 40, QSizePolicy.Minimum, QSizePolicy.Fixed))
        layout.addWidget(darkLabel)
        layout.addWidget(self.darkButton, alignment=Qt.AlignCenter)
        layout.addItem(QSpacerItem(0, 0, QSizePolicy.Minimum, QSizePolicy.Expanding))
        self.setLayout(layout)

    def done(self):
        if self.darkButton.isChecked():
            self.core.setGlobalIconsColor("tray-icon", "#404040")
        elif self.lightButton.isChecked():
            self.core.setGlobalIconsColor("tray-icon", "#8f8f8f")


class PageStyle(Page):
    def __init__(self, core):
        super().__init__(core)
        self.setSubTitle("Select a frame style")

        minimalLabel = Label(bold=True, center=True, text="\nMinimal style")
        minimalDescLabel = Label(
            center=True, text="No toolbar, large resizing corners and more space for text"
        )
        toolbarLabel = Label(bold=True, center=True, text="\nToolbar style")
        toolbarDescLabel = Label(center=True, text="Customizable toolbar with smaller resizing corners")

        self.minimalButton = CheckableButton(RootDirs.WIZARD / "style_minimal.png")
        self.toolbarButton = CheckableButton(RootDirs.WIZARD / "style_toolbar.png")

        layout = QtWidgets.QGridLayout()
        layout.addItem(QSpacerItem(0, 0, QSizePolicy.Minimum, QSizePolicy.Expanding), 0, 0, 1, 1)
        layout.addWidget(self.minimalButton, 1, 0, 1, 1, alignment=Qt.AlignCenter)
        layout.addWidget(self.toolbarButton, 1, 1, 1, 1, alignment=Qt.AlignCenter)
        layout.addWidget(minimalLabel, 2, 0, 1, 1)
        layout.addWidget(toolbarLabel, 2, 1, 1, 1)
        layout.addWidget(minimalDescLabel, 3, 0, 1, 1)
        layout.addWidget(toolbarDescLabel, 3, 1, 1, 1)
        layout.addItem(QSpacerItem(0, 0, QSizePolicy.Minimum, QSizePolicy.Expanding), 4, 0, 1, 1)
        self.setLayout(layout)

    def done(self):
        if self.minimalButton.isChecked():
            self.sdb["profile default"]["style"] = "minimal.css"
        elif self.toolbarButton.isChecked():
            self.sdb["profile default"]["style"] = "toolbar.css"


class PageMode(Page):
    def __init__(self, core):
        super().__init__(core)
        self.setSubTitle("Select a text editing mode")

        htmlLabel = Label(bold=True, center=True, text="\nRich text mode")
        htmlDescLabel = Label(center=True, text="Enable basic text decorations")
        plainLabel = Label(bold=True, center=True, text="\nPlain text mode")
        plainDescLabel = Label(center=True, text="Less featured, lighter and faster")
        vimLabel = Label(bold=True, center=True, text="\nVim mode")
        vimDescLabel = Label(text="Requires Linux, Vim and QTermWidget")

        self.htmlButton = CheckableButton(RootDirs.WIZARD / "mode_rich.png")
        self.plainButton = CheckableButton(RootDirs.WIZARD / "mode_plain.png")
        self.vimButton = CheckableButton(RootDirs.WIZARD / "mode_vim.png")
        self.vimButton.setEnabled("QTermWidget" in sys.modules)

        layout = QtWidgets.QGridLayout()
        layout.addItem(QSpacerItem(0, 0, QSizePolicy.Minimum, QSizePolicy.Expanding), 0, 0, 1, 1)
        layout.addWidget(self.htmlButton, 1, 0, 1, 1, alignment=Qt.AlignCenter)
        layout.addWidget(self.plainButton, 1, 1, 1, 1, alignment=Qt.AlignCenter)
        layout.addWidget(self.vimButton, 1, 3, 1, 1, alignment=Qt.AlignCenter)
        layout.addWidget(htmlLabel, 2, 0, 1, 1)
        layout.addWidget(plainLabel, 2, 1, 1, 1)
        layout.addWidget(vimLabel, 2, 3, 1, 1)
        layout.addWidget(htmlDescLabel, 3, 0, 1, 1)
        layout.addWidget(plainDescLabel, 3, 1, 1, 1)
        layout.addWidget(vimDescLabel, 3, 3, 1, 1)
        layout.addItem(QSpacerItem(0, 0, QSizePolicy.Minimum, QSizePolicy.Expanding), 4, 0, 1, 1)
        self.setLayout(layout)

    def done(self):
        if self.htmlButton.isChecked():
            self.sdb["profile default"]["mode"] = "html"
        elif self.plainButton.isChecked():
            self.sdb["profile default"]["mode"] = "plain"
        elif self.vimButton.isChecked():
            self.sdb["profile default"]["mode"] = "vim"


class PageLinux(Page):
    def __init__(self, core):
        super().__init__(core)
        self.setSubTitle("You're all set !")
        doneLabel = Label(
            10, wrap=True, text="All done ! You may now right click on the tray icon to begin."
        )
        pixmap = QtGui.QPixmap(str(RootDirs.WIZARD / "tray_linux.png"))
        trayImgLabel = QtWidgets.QLabel()
        trayImgLabel.setPixmap(pixmap)
        trayImgLabel.setAlignment(Qt.AlignCenter)
        trayTipLabel = Label(10)
        trayTipLabel.setText(
            "<html><head/><body><p>No taskbar? Open the main menu from the command line :</p>"
            f'<p><span style="font-weight:600;">{ID} -c menu</span></p>'
            "</body></html>"
        )

        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(doneLabel)
        layout.addItem(HSpacer(20, QSizePolicy.Fixed))
        layout.addWidget(trayImgLabel)
        layout.addItem(HSpacer(40, QSizePolicy.Fixed))
        layout.addWidget(trayTipLabel)
        layout.addItem(HSpacer(40, QSizePolicy.Expanding))
        self.setLayout(layout)


class PageWindows(Page):
    def __init__(self, core):
        super().__init__(core)
        self.setSubTitle("You're all set !")

        doneLabel = Label(10, wrap=True)
        doneLabel.setScaledContents(True)
        doneLabel.setText(
            "<html><head/><body><p>All done ! You may now right click on the tray icon to begin.</p><hr/>"
            "<p>It is recommended to set the taskbar icon as always shown. To do so, follow these simple steps :"
            "</p></body></html>"
        )

        step1Label = Label(wrap=True)
        step1Label.setText(
            '<html><head/><body><p><span style="font-size:10pt;">1. Press the </span>'
            '<span style="font-size:10pt; font-weight:600;">Manage taskbar</span>'
            '<span style="font-size:10pt;"> button</span></p></body></html>'
        )

        step2Label = Label(wrap=True)
        if platform.release() == "7":
            step2Path = "tray_w7_1.png"
            step3Path = "tray_w7_2.png"
            step2Label.setText(
                '<html><head/><body><p><span style="font-size:10pt;">2. Uncheck </span>'
                '<span style="font-size:10pt; font-weight:600;">Always show all icons and '
                "notifications</span></p></body></html>"
            )
        else:
            step2Path = "tray_w10_1.png"
            step3Path = "tray_w10_2.png"
            step2Label.setText(
                '<html><head/><body><p><span style="font-size:10pt;">2. Under the </span>'
                '<span style="font-size:10pt; font-weight:600;">Notification area</span>'
                '<span style="font-size:10pt;"> category, click on </span>'
                '<span style="font-size:10pt; font-weight:600;">Select which icons appear '
                "on the taskbar</span></p></body></html>"
            )

        step2Pixmap = QtGui.QPixmap(str(RootDirs.WIZARD / step2Path))
        step2ImgLabel = QtWidgets.QLabel()
        step2ImgLabel.setPixmap(step2Pixmap)

        step3Pixmap = QtGui.QPixmap(str(RootDirs.WIZARD / step3Path))
        step3Label = Label(wrap=True)
        step3Label.setText(
            '<html><head/><body><p><span style="font-size:10pt;">3. Find and enable the </span>'
            f'<span style="font-size:10pt; font-weight:600;">{APP_NAME}</span>'
            '<span style="font-size:10pt;"> icon</span></p></body></html>'
        )

        step3ImgLabel = QtWidgets.QLabel()
        step3ImgLabel.setPixmap(step3Pixmap)

        sizePolicy = QSizePolicy(QSizePolicy.Maximum, QSizePolicy.Fixed)
        trayButton = TrayButton()
        trayButton.setFocusPolicy(Qt.NoFocus)
        trayButton.setSizePolicy(sizePolicy)

        spacer = QSpacerItem(20, 0, QSizePolicy.Fixed, QSizePolicy.Minimum)
        layout = QtWidgets.QGridLayout()
        layout.addWidget(doneLabel, 0, 0, 1, 5)
        layout.addItem(HSpacer(20, QSizePolicy.Expanding), 3, 0, 1, 1)
        layout.addWidget(step1Label, 2, 0, 1, 1)
        layout.addItem(HSpacer(40, QSizePolicy.Expanding), 1, 0, 1, 1)
        layout.addWidget(trayButton, 2, 2, 1, 1)
        layout.addItem(HSpacer(20, QSizePolicy.Expanding), 5, 0, 1, 1)
        layout.addWidget(step2Label, 4, 0, 1, 1)
        layout.addWidget(step2ImgLabel, 4, 2, 1, 1)
        layout.addItem(spacer, 2, 1, 1, 1)
        layout.addWidget(step3Label, 6, 0, 1, 1)
        layout.addWidget(step3ImgLabel, 6, 2, 1, 1)
        self.setLayout(layout)


class Wizard(QtWidgets.QWizard):
    def __init__(self, core):
        super().__init__()
        self.core = core
        self.setWindowTitle(f"{APP_NAME} initial configuration wizard")
        self.resize(780, 460)
        self.setWizardStyle(self.ModernStyle)
        self.accepted.connect(self._postInstall)
        self.rejected.connect(self._postInstall)

        pixmap = core.icons["settings"].pixmap(QtCore.QSize(32, 32))
        self.setPixmap(self.LogoPixmap, pixmap)
        self.button(self.NextButton).clicked.connect(self._next)

        self.addPage(PageHome(core))
        self.addPage(PagePalette(core))
        self.addPage(PageMenuColor(core))
        self.addPage(PageTrayColor(core))
        self.addPage(PageStyle(core))
        self.addPage(PageMode(core))
        if sys.platform.startswith("win"):
            self.addPage(PageWindows(core))
        else:
            self.addPage(PageLinux(core))

    def _createStartupLink(self):
        from win32com.shell import shell, shellcon

        commonLink = (
            Path(shell.SHGetFolderPath(0, shellcon.CSIDL_COMMON_PROGRAMS, 0, 0)) / f"{APP_NAME}.lnk"
        )
        userLink = Path(shell.SHGetFolderPath(0, shellcon.CSIDL_PROGRAMS, 0, 0)) / f"{APP_NAME}.lnk"
        startupLink = Path(shell.SHGetFolderPath(0, shellcon.CSIDL_STARTUP, 0, 0)) / f"{APP_NAME}.lnk"

        nsisLink = commonLink if commonLink.exists() else userLink
        try:
            shutil.copy(nsisLink, startupLink)
        except Exception:
            log.exception(f"Failed to create startup link '{startupLink}'")

    def _next(self):
        """Calls done() method on the previous (now completed) page"""
        last = self.currentId() - 1
        self.page(last).done()

    def _postInstall(self):
        Legacy.prompt(self.core)
        self._startupLinkPrompt()

    def _startupLinkPrompt(self):
        if sys.platform.startswith("win"):
            msg = f"Would you like to launch {APP_NAME}\non Windows startup ?"
            dialog = ConfirmPrompt("Launch on Startup", msg)
            if dialog.exec_() == dialog.Yes:
                self._createStartupLink()



#!/usr/bin/python3
from pathlib import Path
from PyQt5 import QtCore

try:
    from QTermWidget import QTermWidget
except ImportError:
    QTermWidget = QtCore.QObject

try:
    from ..backend import UserDirs, UserFiles, RootDirs, RootFiles, logger
    from ..notes import Note
    from ..notes.console import Terminal
    from ..menus.bars import WordsCounter
except (ValueError, ImportError):
    from backend import UserDirs, UserFiles, RootDirs, RootFiles, logger
    from notes import Note
    from notes.console import Terminal
    from menus.bars import WordsCounter

log = logger.new(__name__)


class Vim(Note):
    mode = "vim"

    def __init__(self, core, path: Path):
        super().__init__(core, path)
        self._initNoteWindow(path)
        self.ref = f"note_{id(self)}"
        self.server = VimServer(self)
        self.body = VimTerminal(self)
        self.gridLayout.addWidget(self.body, 1, 0, 1, 3)
        self.wordsCounter = WordsCounter(self)
        self.msgbarLayout.addWidget(self.wordsCounter)

    def drop(self, mime):
        """Handler for dropped text"""
        with open(self.path, "a", encoding="utf-8", newline="\n") as f:
            f.write(mime.data)


class VimTerminal(Terminal):
    def __init__(self, note):
        super().__init__(note)
        runtime = UserDirs.VIM if UserFiles.VIM.is_file() else RootDirs.VIM
        vimrc = UserFiles.VIM if UserFiles.VIM.is_file() else RootFiles.VIM
        args = [
            "--clean",
            "--noplugin",
            "--servername",
            note.ref,
            "--cmd",
            f"let &rtp.=',{runtime}'",
            "-u",
            str(vimrc),
            str(note.path),
        ]
        self.setShellProgram("vim")
        self.setArgs(args)
        self.startShellProgram()


class VimServer(QtCore.QProcess):
    def __init__(self, note):
        super().__init__()
        self.note = note
        self.setProgram("vim")
        self.setStandardOutputFile(self.nullDevice())
        self.setStandardErrorFile(self.nullDevice())

    def move(self, new: Path):
        """Updates the current file path and working directory in Vim"""
        buffer = f"cd {new.parent} | edit {new.name}"
        autocmd = "autocmd TextChanged,TextChangedI <buffer> silent write"
        args = ["--servername", self.note.ref, "--remote-expr", f'execute("{buffer} | {autocmd}")']

        log.info(f"{self.note.id} : vim {' '.join(args)}")
        self.setArguments(args)
        self.startDetached()



#!/usr/bin/python3
import sys
from pathlib import Path
from PyQt5.QtWidgets import QPlainTextEdit

try:
    from ...backend import logger
except (ValueError, ImportError):
    from backend import logger

log = logger.new(__name__)


def isInstalled() -> bool:
    """Verifies if Antidote is installed"""
    return Path(getBinaryPath()).is_file()


def getBinaryPath() -> str:
    """Finds the path of Antidote executable"""

    def hklm(path: str, name: str) -> str:
        import win32api
        import win32con

        key = win32api.RegOpenKey(win32con.HKEY_LOCAL_MACHINE, path)
        return win32api.RegQueryValueEx(key, name)[0]

    if sys.platform.startswith("win"):
        try:
            folder = hklm(r"Software\Druide informatique inc.\Antidote", "DossierAntidote")
            return rf"{folder}\Antidote.exe"  # MS Windows
        except ImportError:
            log.error("pywin32 module not found")
        except BaseException:
            pass
    else:
        if Path("/usr/local/bin/AgentConnectix").is_file():
            return "/usr/local/bin/AgentConnectix"  # Antidote 10
        elif Path("/usr/local/bin/Antidote").is_file():
            return "/usr/local/bin/Antidote"  # Antidote 9
    return ""  # Unsupported


def getHandler(uid: str, body: QPlainTextEdit, silent=False) -> object:
    """Returns the right handler class according to the user OS"""
    path = getBinaryPath()
    if Path(path).is_file():
        if sys.platform.startswith("win"):
            try:
                from plugins.antidote.com import Antidote
            except (ValueError, ImportError):
                from ...plugins.antidote.com import Antidote

        else:
            try:
                from plugins.antidote.dbus import Antidote
            except (ValueError, ImportError):
                from ...plugins.antidote.dbus import Antidote

        return Antidote(uid, path, body)
    elif silent:
        log.debug("Could not find Antidote binary")
    else:
        raise FileNotFoundError("Could not find Antidote binary")
    return None



#!/usr/bin/python3
import sys
from pathlib import Path
from typing import Tuple
from PyQt5 import QtCore, QtWidgets

try:
    from ..backend import UserDirs, Tuples, logger
except (ValueError, ImportError):
    from backend import UserDirs, Tuples, logger

log = logger.new(__name__)


class ConfirmPrompt(QtWidgets.QMessageBox):
    def __init__(self, title: str, msg: str):
        super().__init__()
        self.setWindowFlags(self.windowFlags() | QtCore.Qt.WindowStaysOnTopHint)
        self.setStandardButtons(self.Yes | self.No)
        self.setDefaultButton(self.No)
        self.setIcon(self.Warning)
        self.setWindowTitle(title)
        self.setText(msg)


def sortQActions(data: list) -> list:
    """Sorts a list of QMenus or QActions"""
    try:
        return sorted(data, key=lambda item: item.title().lower())
    except AttributeError:
        return sorted(data, key=lambda item: item.text().lower())


class AbstractNoteAction(QtWidgets.QAction):
    def __init__(self, parent: QtWidgets.QMenu, path: Path):
        super().__init__(parent)
        self.core = parent.core
        self.path = path
        self.setCheckable(True)
        self.triggered.connect(self._triggered)
        self._setIcon()

    def truncate(self, label: str) -> str:
        """Truncates label if longer than threshold value"""
        threshold = self.core.sdb["general"]["truncate threshold"]
        if len(label) > threshold:
            return f"{label[:threshold - 4]} ..."
        return label

    def _setIcon(self):
        """Sets an icon according to note type"""
        nid = str(self.path.relative_to(UserDirs.NOTES))
        try:
            mode = self.core.pdb[nid]["mode"]
            prefix = mode if mode in ("console", "html", "image") else "plain"
        except KeyError:
            prefix = "image" if self.path.suffix == ".png" else "plain"

        favorites = self.core.ndb["favorites"]
        suffix = "starred" if nid in favorites else "regular"
        icon = self.core.icons[f"{prefix}_{suffix}"]
        self.setIcon(icon)


class CoreAction(QtWidgets.QAction):
    def __init__(self, core, action: Tuples.Action, path: Path = None):
        super().__init__(core)
        self.path = path
        self.action = action
        self.setIcon(action.icon)
        self.setText(action.label)
        self.triggered.connect(self._triggered)

    def _triggered(self):
        """Handles left click event, calls an action"""
        if self.path:
            log.info(f"Core : {self.action.label} : {self.path}")
            self.action.call(self.path)
        else:
            log.info(f"Core : {self.action.label}")
            self.action.call()


class CoreMenu(QtWidgets.QMenu):
    def __init__(self, core):
        super().__init__()
        self.core = core
        self.icons = core.icons
        self.aboutToShow.connect(self._refresh)

        # Preload menu to prevent odd first appearance on xfce, cinnamon, sway, ...
        if not sys.platform.startswith("win"):
            self.popup(QtCore.QPoint(0, 0))
            self.hide()

    def addFoldersList(self):
        """Adds root level folders top menus"""
        folders = [f for f in UserDirs.NOTES.iterdir() if f.is_dir()]
        folders = [SubMenu(self, f) for f in folders]
        for f in sortQActions(folders):
            self.addMenu(f)

    def addNotesList(self):
        """Inserts QActions for locals (root) and loaded sub-notes"""

        def hasParents(path: Path) -> bool:
            """Verifies if a note is located at the root of the notes directory"""
            parents = path.relative_to(UserDirs.NOTES)
            return bool(parents.parts[:-1])

        notes = [self.RootNoteAction(self, f) for f in UserDirs.NOTES.iterdir() if f.is_note()]
        for item in sortQActions(notes):
            self.addAction(item)

        subnotes = [self.RootNoteAction(self, f) for f in self.core.loaded if hasParents(f)]
        for item in sortQActions(subnotes):
            self.addAction(item)

        if not notes and not subnotes:
            item = QtWidgets.QAction(self)
            item.setText("Note folder is empty")
            item.setEnabled(False)
            self.addAction(item)

    def _refresh(self):
        """Updates core menu on request"""
        self.clear()
        for key in self.core.sdb["core menus"]["tray"]:
            if key == "separator":
                self.addSeparator()
            elif key == "folders list":
                self.addFoldersList()
            elif key == "notes list":
                self.addNotesList()
            else:
                try:
                    action = self.core.actions.tray[key]
                    action = CoreAction(self.core, action)
                    self.addAction(action)
                except KeyError:
                    log.exception(f"Invalid action '{key}'")

    class RootNoteAction(AbstractNoteAction):
        def __init__(self, parent: QtWidgets.QMenu, path: Path):
            super().__init__(parent, path)
            if self.path in self.core.loaded:
                isVisible = self.core.loaded[path].isVisible()
                self.setChecked(isVisible)

            label = self.path.stem
            if self.path.is_nested():
                label = self._shortPath(label)
            self.setText(self.truncate(label))

        def _shortPath(self, stem: str) -> str:
            """Abbreviates folder paths when a label exceed truncation threshold"""
            threshold = self.core.sdb["general"]["truncate threshold"]
            parts = self.path.relative_to(UserDirs.NOTES).parts
            label = " / ".join(parts[:-1]) + f" / {stem}"
            if len(label) > threshold:
                label = f"{parts[0]} // {stem}"
            return label

        def _triggered(self):
            """Handles left click event, toggles a note"""
            self.core.notes.toggle(self.path)


class SubMenu(QtWidgets.QMenu):
    def __init__(self, parent: QtWidgets.QMenu, path: Path):
        super().__init__(parent)
        self.core = parent.core
        self.path = path
        self.setTitle(path.name)
        self.aboutToShow.connect(self._refresh)

        loaded = self._loadedCount(path)
        title = f"{path.name} ({loaded})" if loaded else f"{path.name}"
        icon = "folder_active" if loaded else "folder_inactive"
        icon = self.core.icons[icon]
        self.setTitle(title)
        self.setIcon(icon)

    def _addSubActions(self, actions: Tuple):
        """Adds core actions"""
        for key in actions:
            if key == "separator":
                self.addSeparator()
            else:
                try:
                    action = self.core.actions.browser[key]
                    action = CoreAction(self.core, action, self.path)
                    self.addAction(action)
                except KeyError:
                    log.exception(f"Invalid action '{key}'")

    def _loadedCount(self, path: Path) -> int:
        """Counts how many notes of a folder are currently loaded"""
        count = 0
        for f in self.core.loaded:
            if f.is_relative_to(self.path):
                count += 1
        return count

    def _refresh(self):
        """Updates sub-menus on request"""
        self.clear()
        folders = [f for f in self.path.iterdir() if f.is_dir()]
        folders = [SubMenu(self, f) for f in folders]
        for menu in sortQActions(folders):
            self.addMenu(menu)

        files = [f for f in self.path.iterdir() if f.is_note()]
        files = [self.SubNoteAction(self, f) for f in files]
        for note in sortQActions(files):
            self.addAction(note)

        if files or folders:
            actions = self.core.sdb["core menus"]["browser"]
        else:
            actions = ("new", "rename", "move", "open", "separator", "delete")
        self._addSubActions(actions)

    class SubNoteAction(AbstractNoteAction):
        def __init__(self, parent: QtWidgets.QMenu, path: Path):
            super().__init__(parent, path)
            self.setChecked(self.path in self.core.loaded)
            self.setText(self.truncate(path.stem))

        def _triggered(self):
            """Handles left click event, toggles a sub-note"""
            if self.path in self.core.loaded:
                self.core.loaded[self.path].close()
            else:
                self.core.notes.add(self.path)



#!/usr/bin/python3
import re
from pathlib import Path
from typing import Callable, Iterable, Union
from PyQt5 import QtCore, QtGui, QtWidgets

try:
    from ..__id__ import APP_NAME
    from ..backend import UserDirs, logger
    from ..backend.cryptography import Cipher
except (ValueError, ImportError):
    from __id__ import APP_NAME
    from backend import UserDirs, logger
    from backend.cryptography import Cipher

log = logger.new(__name__)


class Search(QtWidgets.QWidget):
    """Self-contained class for the 'find & replace' dialog (text modes)"""

    def __init__(self, core):
        super().__init__()
        self.core = core
        self.note = None
        self.attached = set()
        self._setupUI()

    def closeEvent(self, event: QtGui.QCloseEvent):
        self.hide()

    def eventFilter(self, obj, event):
        if event.type() == QtCore.QEvent.KeyPress:
            if event.key() == QtCore.Qt.Key_Escape:
                self.hide()
        return QtCore.QObject.event(obj, event)

    def hideEvent(self, event: QtGui.QHideEvent):
        self._activateNoteWindow()
        for path in self.attached.copy():
            note = self.core.loaded[path]
            self.note = note
            self._clearHighlighting()
        self.attached.clear()
        super().hideEvent(event)

    def open(self):
        selection = self.body.textCursor().selectedText()
        self.replaceLine.clear()
        self.findLine.setText(selection)
        self.findLine.selectAll()
        self.findLine.setFocus(True)
        self.show()
        self.activateWindow()

    def reattach(func: Callable) -> Callable:
        """Fallback to the next available note when the attached note becomes unavailable"""

        def wrapper(self, *args, **kwargs) -> any:
            try:
                return func(self, *args, **kwargs)
            except RuntimeError:
                self._attachNext()

        return wrapper

    def attach(self, path: Path):
        """Links the search utility with the latest shown note"""
        note = self.core.loaded[path]
        if note.mode in ("plain", "html"):
            self.note = note
            self.attached.add(path)
            self.setWindowTitle(f"Search in '{note.id}'")

    def detach(self, path: Path):
        """Removes the latest hidden note from the attached set"""
        self.attached.discard(path)
        if not self.attached:
            self.hide()

    @property
    def body(self) -> Union[QtWidgets.QTextEdit, QtWidgets.QPlainTextEdit]:
        """Returns the body object (Q_TextEdit) of the currently attached note"""
        if self.note:
            return self.note.body
        return QtWidgets.QPlainTextEdit()

    @reattach
    def find(self, *args):
        """Finds next occurence in text"""
        self._clearHighlighting()
        if not self._findNext() and self.wrapBox.isChecked():
            self.body.moveCursor(QtGui.QTextCursor.Start)
            self._findNext()

    @reattach
    def findAll(self, text: str = None):
        """Highlights all occurences in text"""
        self._clearHighlighting()
        app = self.core.app
        pos = self.body.verticalScrollBar().value()
        extraSelections = self.body.extraSelections()
        self.body.moveCursor(QtGui.QTextCursor.Start)
        while self._findNext(text):
            bg = app.palette().highlight().color().lighter(125)
            fg = app.palette().highlightedText().color().darker(125)
            extra = QtWidgets.QTextEdit.ExtraSelection()
            extra.cursor = self.body.textCursor()
            extra.format.setBackground(bg)
            extra.format.setForeground(fg)
            extraSelections.append(extra)
        self.body.setExtraSelections(extraSelections)
        self.body.verticalScrollBar().setValue(pos)
        self.body.textCursor().clearSelection()
        self._activateNoteWindow()

    @reattach
    def replace(self, *args):
        """Replaces next occurence in text"""
        matchCase = self.caseBox.isChecked()
        selection = self.body.textCursor().selectedText()
        find = self.findLine.text()
        validSelection = (selection == find) if matchCase else (selection.lower() == find.lower())
        if validSelection:
            subs = self.replaceLine.text()
            self.body.textCursor().insertText(subs)
        self.find()

    @reattach
    def replaceAll(self, *args):
        """Replaces all occurences in text"""
        find = self.findLine.text()
        subs = self.replaceLine.text()
        pos = self.body.verticalScrollBar().value()

        self.body.moveCursor(QtGui.QTextCursor.Start)
        cursor = self.body.textCursor()
        cursor.beginEditBlock()
        while self._findNext(find):
            cursor = self.body.textCursor()
            end = cursor.position()
            start = end - len(find)
            cursor.setPosition(start)
            cursor.setPosition(end, QtGui.QTextCursor.KeepAnchor)
            cursor.insertText(subs)
        cursor.endEditBlock()
        self.body.verticalScrollBar().setValue(pos)
        self.findAll(text=subs)  # Highlight the newly replaced text

    def _activateNoteWindow(self):
        """Tries to call activateWindow() on the current or previous attached note"""
        try:
            self.note.activateWindow()
        except RuntimeError:
            self._attachNext()
            self.note.activateWindow()

    def _attachNext(self):
        """Attaches the next available note to the search utility"""
        while self.attached:
            try:
                path = self.attached.pop()
                note = self.core.loaded[path]
                if note.isVisible():
                    self.attach(path)
                    break
            except (KeyError, RuntimeError):
                pass
        else:
            self.hide()

    @reattach
    def _clearHighlighting(self):
        """Removes all highlighting of the current body object"""
        extraSelections = self.body.extraSelections()
        extraSelections.clear()
        self.body.setExtraSelections(extraSelections)

    @reattach
    def _findNext(self, text: str = None) -> bool:
        """Returns the next occurence matching the selected search flags"""
        text = self.findLine.text() if not text else text
        flags = QtGui.QTextDocument.FindFlags()
        flags = flags | QtGui.QTextDocument.FindCaseSensitively if self.caseBox.isChecked() else flags
        flags = flags | QtGui.QTextDocument.FindWholeWords if self.wholeBox.isChecked() else flags
        return self.body.find(text, flags)

    def _setupUI(self):
        self.findLine = QtWidgets.QLineEdit()
        self.findButton = QtWidgets.QPushButton("Find next")
        self.findAllButton = QtWidgets.QPushButton("Find all")
        self.replaceLine = QtWidgets.QLineEdit()
        self.replaceButton = QtWidgets.QPushButton("Replace")
        self.replaceAllButton = QtWidgets.QPushButton("Replace all")
        self.wholeBox = QtWidgets.QCheckBox("Whole Words")
        self.caseBox = QtWidgets.QCheckBox("Match Case")
        self.wrapBox = QtWidgets.QCheckBox("Wrap Search")

        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.addWidget(self.caseBox)
        self.horizontalLayout.addWidget(self.wholeBox)
        self.horizontalLayout.addWidget(self.wrapBox)

        self.layout = QtWidgets.QGridLayout()
        self.layout.addWidget(self.findLine, 0, 0, 1, 1)
        self.layout.addWidget(self.replaceLine, 1, 0, 1, 1)
        self.layout.addWidget(self.findButton, 0, 1, 1, 1)
        self.layout.addWidget(self.replaceButton, 1, 1, 1, 1)
        self.layout.addWidget(self.findAllButton, 0, 2, 1, 1)
        self.layout.addWidget(self.replaceAllButton, 1, 2, 1, 1)
        self.layout.addLayout(self.horizontalLayout, 2, 0, 1, 3)
        self.setLayout(self.layout)

        self.findLine.setTextMargins(5, 1, 5, 1)
        self.replaceLine.setTextMargins(5, 1, 5, 1)
        self.findLine.setPlaceholderText("Search ...")
        self.replaceLine.setPlaceholderText("Replace with ...")
        self.wrapBox.setChecked(True)

        self.findLine.returnPressed.connect(self.findButton.click)
        self.replaceLine.returnPressed.connect(self.replaceButton.click)
        self.findButton.clicked.connect(self.find)
        self.findAllButton.clicked.connect(lambda: self.findAll())
        self.replaceButton.clicked.connect(self.replace)
        self.replaceAllButton.clicked.connect(self.replaceAll)

        self.installEventFilter(self)
        self.setWindowFlags(self.windowFlags() | QtCore.Qt.WindowStaysOnTopHint)
        self.resize(420, 100)


class Launcher(QtWidgets.QWidget):
    """Main class for the notes repository search utility"""

    def __init__(self, core):
        super().__init__()
        self.core = core
        self.timer = QtCore.QTimer(interval=500, singleShot=True)
        self.timer.timeout.connect(self.update)
        self.worker = WorkerThread(core)
        self.workerThread = QtCore.QThread()
        self.worker.moveToThread(self.workerThread)
        self.worker.addItem.connect(self._addItem)
        self.workerThread.start()
        self._setupUI()

    def activate(self, item: QtWidgets.QTreeWidgetItem):
        """Loads the selected note"""
        path = UserDirs.NOTES / item.text(0)
        if not path.is_file():
            log.error(f"File not found '{path}'")
            return
        try:
            note = self.core.loaded[path]
            note.show()
            note.activateWindow()
        except KeyError:
            self.core.notes.add(path)

        if self.hideBox.isChecked():
            self.hide()
        self._jump(path)

    def closeEvent(self, event: QtGui.QCloseEvent):
        self.hide()

    def showEvent(self, event: QtGui.QShowEvent):
        self._moveToCenter()
        self.lineEdit.setFocus(True)
        self.lineEdit.selectAll()
        super().showEvent(event)

    def update(self):
        """Initiates a new search"""
        pattern = self.lineEdit.text()
        if pattern:
            self.listWidget.clear()
            pattern = QtCore.Q_ARG(str, pattern)
            case = QtCore.Q_ARG(bool, self.caseBox.isChecked())
            QtCore.QMetaObject.invokeMethod(
                self.worker, "search", QtCore.Qt.QueuedConnection, pattern, case
            )

    def _addItem(self, path: str, hits: str, icon: str):
        """Appends the items to tree as they are ready"""
        icon = self.core.icons[icon]
        item = TreeWidgetItem(path, hits, icon)
        self.listWidget.addTopLevelItem(item)

    def _jump(self, path: Path):
        """Jumps to first occurence in text"""
        text = self.lineEdit.text()
        searchWidget = self.core.search
        searchWidget.attach(path)
        searchWidget.findLine.setText(text)
        searchWidget.find()

    def _moveToCenter(self):
        """Place widget at the center of the screen"""
        rect = self.core.screen()
        pos = self.frameGeometry()
        pos.moveCenter(rect.center())
        self.move(pos.x(), pos.y())

    def _setupUI(self):
        self.lineEdit = LineEdit(self.core)
        self.listWidget = TreeWidget(self)
        self.caseBox = Checkbox(self.core, "case", "Match Case")
        self.hideBox = Checkbox(self.core, "hide", "Hide on launch")
        self.caseBox.setFocusPolicy(QtCore.Qt.NoFocus)

        self.lineEdit.hide.connect(self.hide)
        self.lineEdit.focusTreeBottom.connect(self.listWidget.focusBottom)
        self.lineEdit.focusTreeTop.connect(self.listWidget.focusTop)
        self.lineEdit.search.connect(self.update)
        self.lineEdit.textChanged.connect(self.timer.start)
        self.listWidget.activate.connect(self.activate)
        self.listWidget.hide.connect(self.hide)
        self.listWidget.focusLine.connect(self.lineEdit.setFocus)
        self.caseBox.stateChanged.connect(self.update)

        self.layout = QtWidgets.QGridLayout()
        self.layout.addWidget(self.lineEdit, 0, 0, 1, 4)
        self.layout.addWidget(self.listWidget, 1, 0, 1, 4)
        self.layout.addWidget(self.caseBox, 2, 0, 1, 1)
        self.layout.addWidget(self.hideBox, 2, 1, 1, 1)
        self.setLayout(self.layout)

        self.setWindowTitle(f"{APP_NAME} — Search")
        self.setWindowFlags(self.windowFlags() | QtCore.Qt.WindowStaysOnTopHint)
        self.resize(450, 250)


class LineEdit(QtWidgets.QLineEdit):
    focusTreeTop = QtCore.pyqtSignal()
    focusTreeBottom = QtCore.pyqtSignal()
    hide = QtCore.pyqtSignal()
    search = QtCore.pyqtSignal()

    def __init__(self, core):
        super().__init__()
        self.setTextMargins(5, 2, 5, 2)
        self.setPlaceholderText("Search for a note ...")

        icon = core.icons["search"]
        self.action = QtWidgets.QAction(icon, "Search")
        self.action.triggered.connect(self.search.emit)
        self.addAction(self.action, QtWidgets.QLineEdit.TrailingPosition)

    def keyPressEvent(self, event: QtGui.QKeyEvent):
        if event.key() in (QtCore.Qt.Key_Return, QtCore.Qt.Key_Enter):
            self.search.emit()

        elif event.key() == QtCore.Qt.Key_Escape:
            self.hide.emit()

        elif event.key() == QtCore.Qt.Key_Up:
            self.focusTreeBottom.emit()

        elif event.key() == QtCore.Qt.Key_Down:
            self.focusTreeTop.emit()

        super().keyPressEvent(event)


class TreeWidget(QtWidgets.QTreeWidget):
    activate = QtCore.pyqtSignal(QtWidgets.QTreeWidgetItem)
    focusLine = QtCore.pyqtSignal(bool)
    hide = QtCore.pyqtSignal()

    def __init__(self, launcher):
        super().__init__()
        self.launcher = launcher
        self.itemDoubleClicked.connect(self.activate.emit)
        self.setAlternatingRowColors(True)
        self.setSortingEnabled(True)
        self.setIndentation(5)

        self.setHeaderLabels(["path", "hits"])
        self.setColumnWidth(1, 0)  # Minimizes hits column
        self.sortItems(1, QtCore.Qt.DescendingOrder)  # Sort by hits
        self.header().setSectionResizeMode(0, QtWidgets.QHeaderView.Stretch)
        self.header().setStretchLastSection(False)  # Minimizes hits column
        self.header().hide()

    def focusBottom(self):
        """Selects the last item of the tree"""
        last = self.topLevelItem(self.topLevelItemCount() - 1)
        self.setCurrentItem(last)
        self.setFocus(True)

    def focusTop(self):
        """Selects the first item of the tree"""
        first = self.topLevelItem(0)
        self.setCurrentItem(first)
        self.setFocus(True)

    def keyPressEvent(self, event: QtGui.QKeyEvent):
        if event.key() in (QtCore.Qt.Key_Return, QtCore.Qt.Key_Enter):
            self.activate.emit(self.currentItem())

        elif event.key() == QtCore.Qt.Key_Escape:
            self.hide.emit()

        elif event.key() == QtCore.Qt.Key_Up:
            if self.currentIndex().row() == 0:
                self.focusBottom()
                event.ignore()

        elif event.key() == QtCore.Qt.Key_Down:
            if self.currentIndex().row() == self.topLevelItemCount() - 1:
                self.focusTop()
                event.ignore()
        else:
            self._postKeyEvent(event)
            self.focusLine.emit(True)

        if event.isAccepted():
            super().keyPressEvent(event)

    def _postKeyEvent(self, event: QtGui.QKeyEvent):
        """Forwards key events to the LineEdit widget"""
        try:
            sequence = QtGui.QKeySequence(event.key())
            sequence = sequence.toString().lower()
            sequence = sequence.replace("space", " ")
            sequence.encode("utf-8")
        except UnicodeEncodeError:
            return
        keyEvent = QtGui.QKeyEvent(QtCore.QEvent.KeyPress, event.key(), event.modifiers(), sequence)
        QtCore.QCoreApplication.postEvent(self.launcher.lineEdit, keyEvent)


class TreeWidgetItem(QtWidgets.QTreeWidgetItem):
    def __init__(self, path: str, hits: str, icon: QtGui.QIcon):
        super().__init__()
        self.setTextAlignment(1, QtCore.Qt.AlignCenter)
        self.setIcon(0, icon)
        self.setText(0, path)
        self.setText(1, hits)

    def __lt__(self, other: QtWidgets.QTreeWidgetItem) -> bool:
        """Sorts numbers higher than 10"""
        column = self.treeWidget().sortColumn()
        key1 = self.text(column)
        key2 = other.text(column)
        return int(key1) < int(key2)


class Checkbox(QtWidgets.QCheckBox):
    def __init__(self, core, key: str, text: str):
        super().__init__()
        self.core = core
        self.key = key
        self.setText(text)
        self.setChecked(core.sdb["launcher"][key])
        self.stateChanged.connect(self._stateChanged)

    def _stateChanged(self, state: int):
        state = not (state == QtCore.Qt.Unchecked)
        self.core.sdb["launcher"][self.key] = state


class WorkerThread(QtCore.QObject):
    addItem = QtCore.pyqtSignal(str, str, str)

    def __init__(self, core):
        self.core = core
        super().__init__()

    @QtCore.pyqtSlot(str, bool)
    def search(self, pattern: str, case: bool):
        """Returns all files that match a text pattern"""
        for path, hits in FileSearch.seekContent(self.core, UserDirs.NOTES, pattern, case):
            self.addItem.emit(str(path), str(hits), "plain_regular")

        for path, hits in FileSearch.seekFileName(UserDirs.NOTES, pattern, case):
            self.addItem.emit(str(path), str(hits), "rename")


class FileSearch:
    @classmethod
    def seekContent(cls, core, src: Path, pattern: str, exact=False) -> Iterable:
        """Searches notes directory for matching content"""
        for path in [f for f in src.rglob("*.txt") if f.is_file()]:
            content = cls._toPlainText(path) if cls._isHtml(path) else cls._read(path)
            if content.startswith(Cipher.MAGIC):
                decrypted = path in core.loaded and core.loaded[path].hasClearText()
                content = core.loaded[path].content if decrypted else ""
            hits = content.count(pattern) if exact else cls._lazyCount(content, pattern)
            if hits:
                yield path.relative_to(src), hits

    @classmethod
    def seekFileName(cls, src: Path, pattern: str, exact=False) -> Iterable:
        """Searches notes directory for matching filename"""
        for path in [f.relative_to(src) for f in src.rglob("*.txt") if f.is_file()]:
            hits = path.stem.count(pattern) if exact else cls._lazyCount(path.stem, pattern)
            if hits:
                yield path, hits

    def _isHtml(path: Path) -> bool:
        """Verifies if the note file has HTML content"""
        with open(path, encoding="utf8") as f:
            first = f.readline()
        return first.startswith("<!DOCTYPE HTML")

    def _lazyCount(content: str, pattern: str) -> int:
        """Counts all matching pattern regardless of its case"""
        try:
            found = re.findall(pattern, content, re.IGNORECASE)
        except re.error:
            pattern = rf"\{pattern}"
            found = re.findall(pattern, content, re.IGNORECASE)
        return len(found)

    def _read(path: Path) -> str:
        """Returns a file content"""
        with open(path, encoding="utf8") as f:
            return f.read()

    def _toPlainText(path: Path) -> str:
        """Extracts plain text from an HTML note"""
        with open(path, encoding="utf8") as f:
            content = f.read()
        qtext = QtWidgets.QTextEdit()
        qtext.setHtml(content)
        return qtext.toPlainText()
